from inspect import Traceback
from signal import getsignal, SIG_IGN, SIGINT, signal as signal_, Signals
from types import FrameType
from typing import Type


class DelayedKeyboardInterrupt:

    def __init__(self, in_thread: bool = False) -> None:
        """
        :param in_thread: Whether or not we're living in a thread or not
        """
        self.in_thread = in_thread
        self.signal_received = None

    def __enter__(self) -> None:
        # When we're in a thread we can't use signal handling
        if not self.in_thread:
            self.signal_received = False
            self.old_handler = signal_(SIGINT, self.handler)

    def handler(self, sig: Signals, frame: FrameType) -> None:
        self.signal_received = (sig, frame)

    def __exit__(self, exc_type: Type, exc_val: Exception, exc_tb: Traceback) -> None:
        if not self.in_thread:
            signal_(SIGINT, self.old_handler)
            if self.signal_received:
                self.old_handler(*self.signal_received)


class DisableKeyboardInterruptSignal:

    def __enter__(self) -> None:
        # Prevent signal from propagating to child process
        self._handler = getsignal(SIGINT)
        ignore_keyboard_interrupt()

    def __exit__(self, exc_type: Type, exc_val: Exception, exc_tb: Traceback) -> None:
        # Restore signal
        signal_(SIGINT, self._handler)


def ignore_keyboard_interrupt():
    signal_(SIGINT, SIG_IGN)



import itertools
import multiprocessing as mp
import warnings
from typing import Any, Callable, Iterable, List, Optional, Sized, Tuple, Union

from dataclasses import dataclass, field
from tqdm import tqdm

from mpire.context import RUNNING_WINDOWS, DEFAULT_START_METHOD

# Typedefs
CPUList = List[Union[int, List[int]]]


@dataclass(init=True, frozen=False)
class WorkerPoolParams:
    """
    Data class for all :obj:`mpire.WorkerPool` parameters.
    """
    n_jobs: Optional[int]
    _n_jobs: int = field(init=False, repr=False)
    cpu_ids: Optional[CPUList]
    _cpu_ids: CPUList = field(init=False, repr=False)
    daemon: bool = True
    shared_objects: Any = None
    pass_worker_id: bool = False
    use_worker_state: bool = False
    start_method: str = DEFAULT_START_METHOD
    keep_alive: bool = False
    use_dill: bool = False
    enable_insights: bool = False

    @property
    def n_jobs(self) -> Optional[int]:
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, n_jobs: Optional[int]) -> None:
        self._n_jobs = n_jobs or mp.cpu_count()

    @property
    def cpu_ids(self) -> CPUList:
        return self._cpu_ids

    @cpu_ids.setter
    def cpu_ids(self, cpu_ids: Optional[CPUList]) -> None:
        self._cpu_ids = self._check_cpu_ids(cpu_ids)

    def _check_cpu_ids(self, cpu_ids: Optional[CPUList]) -> CPUList:
        """
        Checks the cpu_ids parameter for correctness

        :param cpu_ids: List of CPU IDs to use for pinning child processes to specific CPUs. The list must be as long as
            the number of jobs used (if ``n_jobs`` equals ``None`` it must be equal to ``mpire.cpu_count()``), or the
            list must have exactly one element. In the former case, element x specifies the CPU ID(s) to use for child
            process x. In the latter case the single element specifies the CPU ID(s) for all child  processes to use. A
            single element can be either a single integer specifying a single CPU ID, or a list of integers specifying
            that a single child process can make use of multiple CPU IDs. If ``None``, CPU pinning will be disabled.
            Note that CPU pinning may only work on Linux based systems
        :return: cpu_ids
        """
        # Check CPU IDs
        converted_cpu_ids = []
        if cpu_ids:
            # Check number of arguments
            if len(cpu_ids) != 1 and len(cpu_ids) != self.n_jobs:
                raise ValueError("Number of CPU IDs (%d) does not match number of jobs (%d)" %
                                 (len(cpu_ids), self.n_jobs))

            # Convert CPU IDs to proper format and find the max and min CPU ID
            max_cpu_id = 0
            min_cpu_id = 0
            for cpu_id in cpu_ids:
                if isinstance(cpu_id, list):
                    converted_cpu_ids.append(cpu_id)
                    max_cpu_id = max(max_cpu_id, max(cpu for cpu in cpu_id))
                    min_cpu_id = min(min_cpu_id, min(cpu for cpu in cpu_id))
                elif isinstance(cpu_id, int):
                    converted_cpu_ids.append([cpu_id])
                    max_cpu_id = max(max_cpu_id, cpu_id)
                    min_cpu_id = min(min_cpu_id, cpu_id)
                else:
                    raise TypeError("CPU ID(s) must be either a list or a single integer")

            # Check max CPU ID
            if max_cpu_id >= mp.cpu_count():
                raise ValueError("CPU ID %d exceeds the maximum CPU ID available on your system: %d" %
                                 (max_cpu_id, mp.cpu_count() - 1))

            # Check min CPU ID
            if min_cpu_id < 0:
                raise ValueError("CPU IDs cannot be negative")

        # If only one item is given, use this item for all child processes
        if len(converted_cpu_ids) == 1:
            converted_cpu_ids = list(itertools.repeat(converted_cpu_ids[0], self.n_jobs))

        return converted_cpu_ids


@dataclass(init=True, frozen=True)
class WorkerMapParams:
    """
    Data class for all :meth:`mpire.WorkerPool.map` parameters that need to be passed on to a worker.
    """
    # User provided functions to call, provided to a map function
    func: Callable
    worker_init: Optional[Callable] = None
    worker_exit: Optional[Callable] = None

    # Number of (chunks of) jobs a child process can process before requesting a restart
    worker_lifespan: Optional[int] = None

    def __eq__(self, other: 'WorkerMapParams') -> bool:
        """
        :param other: Other WorkerMapConfig
        :return: Whether the configs are the same
        """
        if other.worker_init != self.worker_init or other.worker_exit != self.worker_exit:
            warnings.warn("You're changing either the worker_init and/or worker_exit function while keep_alive is "
                          "enabled. Be aware this can have undesired side-effects as worker_init functions are only "
                          "executed when a worker is started and worker_exit functions when a worker is terminated.",
                          RuntimeWarning, stacklevel=2)

        return (other.func == self.func and
                other.worker_init == self.worker_init and
                other.worker_exit == self.worker_exit and
                other.worker_lifespan == self.worker_lifespan)


def check_map_parameters(pool_params: WorkerPoolParams, iterable_of_args: Union[Sized, Iterable],
                         iterable_len: Optional[int], max_tasks_active: Optional[int],
                         chunk_size: Optional[Union[int, float]], n_splits: Optional[int],
                         worker_lifespan: Optional[int], progress_bar: bool, progress_bar_position: int) \
        -> Tuple[Optional[int], int, Optional[int], Union[bool, tqdm]]:
    """
    Check the parameters provided to any (i)map function. Also extracts the number of tasks and can modify the
    ``chunk_size`` and ``progress_bar`` parameters.

    :param pool_params: WorkerPool config
    :param iterable_of_args: A numpy array or an iterable containing tuples of arguments to pass to a worker, which
        passes it to the function
    :param iterable_len: Number of elements in the ``iterable_of_args``. When chunk_size is set to ``None`` it needs
        to know the number of tasks. This can either be provided by implementing the ``__len__`` function on the
        iterable object, or by specifying the number of tasks
    :param max_tasks_active: Maximum number of active tasks in the queue. Use ``None`` to not limit the queue
    :param chunk_size: Number of simultaneous tasks to give to a worker. If ``None`` it will generate ``n_jobs * 4``
        number of chunks
    :param n_splits: Number of splits to use when ``chunk_size`` is ``None``
    :param worker_lifespan: Number of chunks a worker can handle before it is restarted. If ``None``, workers will
        stay alive the entire time. Use this when workers use up too much memory over the course of time
    :param progress_bar: When ``True`` it will display a progress bar
    :param progress_bar_position: Denotes the position (line nr) of the progress bar. This is useful wel using
        multiple progress bars at the same time
    :return: Number of tasks, max tasks active, chunk size, progress bar
    """
    # Get number of tasks
    n_tasks = None
    if iterable_len is not None:
        n_tasks = iterable_len
    elif hasattr(iterable_of_args, '__len__'):
        n_tasks = len(iterable_of_args)
    elif chunk_size is None or progress_bar:
        warnings.warn('Failed to obtain length of iterable when chunk size or number of splits is None and/or a '
                      'progress bar is requested. Chunk size is set to 1 and no progress bar will be shown. '
                      'Remedy: either provide an iterable with a len() function or specify iterable_len in the '
                      'function call', RuntimeWarning, stacklevel=2)
        chunk_size = 1
        progress_bar = False

    # Check chunk_size parameter
    if chunk_size is not None:
        if not isinstance(chunk_size, (int, float)):
            raise TypeError('chunk_size should be either None or an integer value')
        elif chunk_size <= 0:
            raise ValueError('chunk_size should be a positive integer > 0')

    # Check n_splits parameter (only when chunk_size is None)
    else:
        if not (isinstance(n_splits, int) or n_splits is None):
            raise TypeError('n_splits should be either None or an integer value')
        if isinstance(n_splits, int) and n_splits <= 0:
            raise ValueError('n_splits should be a positive integer > 0')

    # Check max_tasks_active parameter
    if max_tasks_active is None:
        max_tasks_active = pool_params.n_jobs * 2
    elif isinstance(max_tasks_active, int):
        if max_tasks_active <= 0:
            raise ValueError('max_tasks_active should be a positive integer or None')
    else:
        raise TypeError('max_tasks_active should be a positive integer or None')

    # If worker lifespan is not None or not a positive integer, raise
    if isinstance(worker_lifespan, int):
        if worker_lifespan <= 0:
            raise ValueError('worker_lifespan should be either None or a positive integer (> 0)')
    elif worker_lifespan is not None:
        raise TypeError('worker_lifespan should be either None or a positive integer (> 0)')

    # Progress bar is currently not supported on Windows when using threading as start method. For reasons still
    # unknown we get a TypeError: cannot pickle '_thread.Lock' object.
    if RUNNING_WINDOWS and progress_bar and pool_params.start_method == "threading":
        raise ValueError("Progress bar is currently not supported on Windows when using start_method='threading'")

    # Progress bar position should be a positive integer
    if not isinstance(progress_bar_position, int):
        raise TypeError('progress_bar_position should be a positive integer (>= 0)')
    if progress_bar_position < 0:
        raise ValueError('progress_bar_position should be a positive integer (>= 0)')

    return n_tasks, max_tasks_active, chunk_size, progress_bar



import collections.abc
try:
    import dill as pickle
except ImportError:
    import pickle
import multiprocessing as mp
import signal
import traceback
import _thread
from datetime import datetime
from functools import partial
from threading import Thread
from typing import Any, Callable, List, Optional, Tuple, Union, Type

try:
    import multiprocess
    DILL_INSTALLED = True
except ImportError:
    DILL_INSTALLED = False
try:
    import numpy as np
    NUMPY_INSTALLED = True
except ImportError:
    np = None
    NUMPY_INSTALLED = False

from mpire.comms import NEW_MAP_PARAMS_PILL, NON_LETHAL_POISON_PILL, POISON_PILL, WorkerComms
from mpire.context import FORK_AVAILABLE, MP_CONTEXTS, RUNNING_WINDOWS
from mpire.dashboard.connection_utils import DashboardConnectionDetails, set_dashboard_connection
from mpire.exception import CannotPickleExceptionError, StopWorker
from mpire.insights import WorkerInsights
from mpire.params import WorkerMapParams, WorkerPoolParams
from mpire.tqdm_utils import TqdmConnectionDetails, TqdmManager
from mpire.utils import TimeIt


class AbstractWorker:
    """
    A multiprocessing helper class which continuously asks the queue for new jobs, until a poison pill is inserted
    """

    def __init__(self, worker_id: int, pool_params: WorkerPoolParams, map_params: WorkerMapParams,
                 worker_comms: WorkerComms, worker_insights: WorkerInsights,
                 tqdm_connection_details: TqdmConnectionDetails,
                 dashboard_connection_details: DashboardConnectionDetails, start_time: datetime) -> None:
        """
        :param worker_id: Worker ID
        :param pool_params: WorkerPool parameters
        :param map_params: WorkerPool map parameters
        :param worker_comms: Worker communication objects (queues, locks, events, ...)
        :param worker_insights: WorkerInsights object which stores the worker insights
        :param tqdm_connection_details: Tqdm manager host, and whether the manager is started/connected
        :param dashboard_connection_details: Dashboard manager host, port_nr and whether a dashboard is
            started/connected
        :param start_time: `datetime` object indicating at what time the Worker instance was created and started
        """
        super().__init__()

        # Parameters
        self.worker_id = worker_id
        self.pool_params = pool_params
        self.map_params = map_params
        self.worker_comms = worker_comms
        self.worker_insights = worker_insights
        self.tqdm_connection_details = tqdm_connection_details
        self.dashboard_connection_details = dashboard_connection_details
        self.start_time = start_time

        # Worker state
        self.worker_state = {}

        # Local variables needed for each worker
        self.progress_bar_last_updated = datetime.now()
        self.progress_bar_n_tasks_completed = 0
        self.max_task_duration_last_updated = datetime.now()
        self.max_task_duration_list = self.worker_insights.get_max_task_duration_list(self.worker_id)

        # Exception handling variables
        if self.pool_params.start_method == 'threading':
            ctx = MP_CONTEXTS['threading']
        else:
            ctx = MP_CONTEXTS['mp_dill' if self.pool_params.use_dill else 'mp'][self.pool_params.start_method]
        self.is_running = False
        self.is_running_lock = ctx.Lock()

        # Register handler for graceful shutdown. This doesn't work on Windows
        if not RUNNING_WINDOWS:
            signal.signal(signal.SIGUSR1, self._exit_gracefully)

    def _exit_gracefully(self, *_) -> None:
        """
        This function is called when the main process sends a kill signal to this process. This can only mean another
        child process encountered an exception which means we should exit.

        If this process is in the middle of running the user defined function we raise a StopWorker exception (which is
        then caught by the ``_run_safely()`` function) so we can quit gracefully.
        """
        # A rather complex locking and exception mechanism is used here so we can make sure we only raise an exception
        # when we should. We want to make sure we only raise an exception when the user function is called. If it is
        # running the exception is caught and `run` will return. If it is running and the user function throws another
        # exception first, it depends on who obtains the lock first. If `run` obtains it, it will set `running` to
        # False, meaning we won't raise and `run` will return. If this function obtains it first it will throw, which
        # again is caught by the `run` function, which will return.
        self.worker_comms.signal_kill_signal_received()
        with self.is_running_lock:
            if self.is_running:
                self.is_running = False
                raise StopWorker

    def _exit_gracefully_windows(self) -> None:
        """
        Windows doesn't fully support signals as Unix-based systems do. Therefore, we have to work around it. This
        function is started in a thread. We wait for a kill signal (Event object) and interrupt the main thread if we
        got it (derived from https://stackoverflow.com/a/40281422). This will raise a KeyboardInterrupt, which is then
        caught by the signal handler, which in turn checks if we need to raise a StopWorker.

        Note: functions that release the GIL won't be interupted by this procedure (e.g., time.sleep). If graceful
        shutdown takes too long the process will be terminated by the main process. If anyone has a better approach for
        graceful interrupt in Windows, please create a PR.
        """
        while self.worker_comms.is_worker_alive(self.worker_id):
            if self.worker_comms.wait_for_exception_thrown(timeout=0.1):
                _thread.interrupt_main()
                return

    def run(self) -> None:
        """
        Continuously asks the tasks queue for new task arguments. When not receiving a poisonous pill or when the max
        life span is not yet reached it will execute the new task and put the results in the results queue.
        """
        # Enable graceful shutdown for Windows. Note that we can't kill threads in Python
        if RUNNING_WINDOWS and self.pool_params.start_method != "threading":
            signal.signal(signal.SIGINT, self._exit_gracefully)
            t = Thread(target=self._exit_gracefully_windows)
            t.start()

        with self.worker_comms.get_worker_dead_lock(self.worker_id):
            self.worker_comms.signal_worker_alive(self.worker_id)

        # Set tqdm and dashboard connection details. This is needed for nested pools and in the case forkserver or
        # spawn is used as start method
        TqdmManager.set_connection_details(self.tqdm_connection_details)
        set_dashboard_connection(self.dashboard_connection_details, auto_connect=False)

        try:
            # Store how long it took to start up
            self.worker_insights.update_start_up_time(self.worker_id, self.start_time)

            # Obtain additional args to pass to the function
            additional_args = []
            if self.pool_params.pass_worker_id:
                additional_args.append(self.worker_id)
            if self.pool_params.shared_objects is not None:
                additional_args.append(self.pool_params.shared_objects)
            if self.pool_params.use_worker_state:
                additional_args.append(self.worker_state)

            # Run initialization function. If it returns True it means an exception occurred and we should exit
            if self.map_params.worker_init and self._run_init_func(additional_args):
                return

            # Determine what function to call. If we have to keep in mind the order (for map) we use the helper function
            # with idx support which deals with the provided idx variable.
            func = self._get_func(additional_args)

            n_tasks_executed = 0
            while self.map_params.worker_lifespan is None or n_tasks_executed < self.map_params.worker_lifespan:

                # Obtain new chunk of jobs
                with TimeIt(self.worker_insights.worker_waiting_time, self.worker_id):
                    next_chunked_args = self.worker_comms.get_task(self.worker_id)

                # Force update task insights and progress bar when we got a (non-lethal) poison pill. At this point, we
                # know for sure that all results have been processed. In case of a lethal pill we additionally run the
                # worker exit function, wait for all the exit results to be obtained, wait for the progress bar to be
                # done, and stop. Otherwise, we simply continue
                if next_chunked_args == POISON_PILL or next_chunked_args == NON_LETHAL_POISON_PILL:
                    self._update_task_insights(force_update=True)
                    self._update_progress_bar(force_update=True)
                    self.worker_comms.task_done(self.worker_id)
                    if next_chunked_args == POISON_PILL:
                        if self.map_params.worker_exit:
                            self._run_exit_func(additional_args)
                            self.worker_comms.wait_until_all_exit_results_obtained()
                        if self.worker_comms.has_progress_bar():
                            self.worker_comms.wait_until_progress_bar_is_complete()
                        return
                    else:
                        continue

                # Update the map parameters of this function when new parameters are provided
                elif next_chunked_args == NEW_MAP_PARAMS_PILL:
                    self.worker_comms.task_done(self.worker_id)
                    self.map_params = self.worker_comms.get_task(self.worker_id)
                    func = self._get_func(additional_args)
                    self.worker_comms.task_done(self.worker_id)
                    continue

                # When we recieved None this means we need to stop because of an exception in the main process
                elif next_chunked_args is None:
                    return

                # Execute jobs in this chunk
                try:
                    results = []
                    for args in next_chunked_args:

                        # Try to run this function and save results
                        results_part, should_return = self._run_func(func, args)
                        if should_return:
                            return
                        results.append(results_part)

                        # Update progress bar info
                        self._update_progress_bar()

                    # Send results back to main process
                    self.worker_comms.add_results(self.worker_id, results)
                    n_tasks_executed += len(results)

                # In case an exception occurred and we need to return, we want to call task_done no matter what
                finally:
                    self.worker_comms.task_done(self.worker_id)

                # Update task insights
                self._update_task_insights()

            # Max lifespan reached
            self._update_task_insights(force_update=True)
            self._update_progress_bar(force_update=True)
            if self.map_params.worker_exit and self._run_exit_func(additional_args):
                return

            # Notify WorkerPool to start a new worker
            if self.map_params.worker_lifespan is not None and n_tasks_executed == self.map_params.worker_lifespan:
                self.worker_comms.signal_worker_restart(self.worker_id)

        finally:
            with self.worker_comms.get_worker_dead_lock(self.worker_id):
                self.worker_comms.signal_worker_dead(self.worker_id)

    def _get_func(self, additional_args: List) -> Callable:
        """
        Determine what function to call. If we have to keep in mind the order (for map) we use the helper function with
        idx support which deals with the provided idx variable.

        :param additional_args: Additional args list
        :return: Function to call
        """
        return partial(self._helper_func_with_idx if self.worker_comms.keep_order() else self._helper_func,
                       partial(self.map_params.func, *additional_args))

    def _run_init_func(self, additional_args: List) -> bool:
        """
        Runs the init function when provided.

        :param additional_args: Additional args to pass to the function (worker ID, shared objects, worker state)
        :return: True when the worker needs to shut down, False otherwise
        """
        def _init_func():
            with TimeIt(self.worker_insights.worker_init_time, self.worker_id):
                self.map_params.worker_init(*additional_args)

        return self._run_safely(_init_func, no_args=True)[1]

    def _run_func(self, func: Callable, args: List) -> Tuple[Any, bool]:
        """
        Runs the main function when provided.

        :param func: Function to call
        :param args: Args to pass to the function
        :return: Tuple containing results from the function and a boolean value indicating whether the worker needs to
            shut down
        """
        def _func():
            with TimeIt(self.worker_insights.worker_working_time, self.worker_id,
                        self.max_task_duration_list, lambda: self._format_args(args, separator=' | ')):
                results = func(args)
            self.worker_insights.update_n_completed_tasks(self.worker_id)
            return results

        return self._run_safely(_func, args)

    def _run_exit_func(self, additional_args: List) -> bool:
        """
        Runs the exit function when provided and stores its results.

        :param additional_args: Additional args to pass to the function (worker ID, shared objects, worker state)
        :return: True when the worker needs to shut down, False otherwise
        """
        def _exit_func():
            with TimeIt(self.worker_insights.worker_exit_time, self.worker_id):
                return self.map_params.worker_exit(*additional_args)

        results, should_return = self._run_safely(_exit_func, no_args=True)

        if should_return:
            return True
        else:
            self.worker_comms.add_exit_results(self.worker_id, results)
            return False

    def _run_safely(self, func: Callable, exception_args: Optional[Any] = None,
                    no_args: bool = False) -> Tuple[Any, bool]:
        """
        A rather complex locking and exception mechanism is used here so we can make sure we only raise an exception
        when we should. See `_exit_gracefully` for more information.

        :param func: Function to run
        :param exception_args: Arguments to pass to `_format_args` when an exception occurred
        :param no_args: Whether there were any args at all
        :return: True when the worker needs to shut down, False otherwise
        """
        if self.worker_comms.exception_thrown():
            return None, True

        try:

            try:
                # Obtain lock and try to run the function. During this block a StopWorker exception from the parent
                # process can come through to signal we should stop
                with self.is_running_lock:
                    self.is_running = True
                results = func()
                with self.is_running_lock:
                    self.is_running = False

            except StopWorker:
                # The main process tells us to stop working, shutting down
                raise

            except (Exception, SystemExit) as err:
                # An exception occurred inside the provided function. Let the signal handler know it shouldn't raise any
                # StopWorker exceptions from the parent process anymore, we got this.
                with self.is_running_lock:
                    self.is_running = False

                # Pass exception to parent process and stop
                self._raise(exception_args, no_args, err)
                raise StopWorker

        except StopWorker:
            # Stop working
            return None, True

        # Carry on
        return results, False

    def _raise(self, args: Any, no_args: bool, err: Union[Exception, SystemExit]) -> None:
        """
        Create exception and pass it to the parent process. Let other processes know an exception is set

        :param args: Funtion arguments where exception was raised
        :param no_args: Whether there were any args at all
        :param err: Exception that should be passed on to parent process
        """
        # Only one process can throw at a time
        with self.worker_comms.exception_lock:

            # Only raise an exception when this process is the first one to raise. We do this because when the first
            # exception is caught by the main process the workers are joined which can cause a deadlock on draining the
            # exception queue. By only allowing one process to throw we know for sure that the exception queue will be
            # empty when the first one arrives.
            if not self.worker_comms.exception_thrown():

                # Let others know we need to stop
                self.worker_comms.signal_exception_thrown()

                # Create traceback string
                traceback_str = "\n\nException occurred in Worker-%d with the following arguments:\n%s\n%s" % (
                    self.worker_id, self._format_args(args, no_args), traceback.format_exc()
                )

                # Sometimes an exception cannot be pickled (i.e., we get the _pickle.PickleError: Can't pickle
                # <class ...>: it's not the same object as ...). We check that here by trying the pickle.dumps manually.
                # The call to `queue.put` creates a thread in which it pickles and when that raises an exception we
                # cannot catch it.
                try:
                    pickle.dumps(type(err))
                except pickle.PicklingError:
                    err = CannotPickleExceptionError()

                # Add exception. When we have a progress bar, we add an additional one
                self.worker_comms.add_exception(type(err), traceback_str)
                if self.worker_comms.has_progress_bar():
                    self.worker_comms.add_exception(type(err), traceback_str)

    def _format_args(self, args: Any, no_args: bool = False, separator: str = '\n') -> str:
        """
        Format the function arguments to a string form.

        :param args: Funtion arguments
        :param no_args: Whether there were any args at all. If not, then return special string
        :param separator: String to use as separator between arguments
        :return: String containing the task arguments
        """
        # Determine function arguments
        func_args = args[1] if args and self.worker_comms.keep_order() else args
        if no_args:
            return "N/A"
        elif isinstance(func_args, dict):
            return separator.join("Arg %s: %s" % (str(key), repr(value)) for key, value in func_args.items())
        elif isinstance(func_args, collections.abc.Iterable) and not isinstance(func_args, (str, bytes)):
            return separator.join("Arg %d: %s" % (arg_nr, repr(arg)) for arg_nr, arg in enumerate(func_args))
        else:
            return "Arg 0: %s" % func_args

    def _helper_func_with_idx(self, func: Callable, args: Tuple[int, Any]) -> Tuple[int, Any]:
        """
        Helper function which calls the function `func` but preserves the order index

        :param func: Function to call each time new task arguments become available
        :param args: Tuple of ``(idx, _args)`` where ``_args`` correspond to the arguments to pass on to the function.
            ``idx`` is used to preserve order
        :return: (idx, result of calling the function with the given arguments) tuple
        """
        return args[0], self._call_func(func, args[1])

    def _helper_func(self, func: Callable, args: Any) -> Any:
        """
        Helper function which calls the function `func`

        :param func: Function to call each time new task arguments become available
        :param args: Arguments to pass on to the function
        :return: Result of calling the function with the given arguments) tuple
        """
        return self._call_func(func, args)

    @staticmethod
    def _call_func(func: Callable, args: Any) -> Any:
        """
        Helper function which calls the function `func` and passes the arguments in the correct way

        :param func: Function to call each time new task arguments become available
        :param args: Arguments to pass on to the function
        :return: Result of calling the function with the given arguments) tuple
        """
        if isinstance(args, dict):
            return func(**args)
        elif (isinstance(args, collections.abc.Iterable) and not isinstance(args, (str, bytes)) and not
              (NUMPY_INSTALLED and isinstance(args, np.ndarray))):
            return func(*args)
        else:
            return func(args)

    def _update_progress_bar(self, force_update: bool = False) -> None:
        """
        Update the progress bar data

        :param force_update: Whether to force an update
        """
        if self.worker_comms.has_progress_bar():
            (self.progress_bar_last_updated,
             self.progress_bar_n_tasks_completed) = self.worker_comms.task_completed_progress_bar(
                self.worker_id, self.progress_bar_last_updated, self.progress_bar_n_tasks_completed, force_update
             )

    def _update_task_insights(self, force_update: bool = False) -> None:
        """
        Update the task insights data

        :param force_update: Whether to force an update
        """
        self.max_task_duration_last_updated = self.worker_insights.update_task_insights(
            self.worker_id, self.max_task_duration_last_updated, self.max_task_duration_list, force_update=force_update
        )


if FORK_AVAILABLE:
    class ForkWorker(AbstractWorker, MP_CONTEXTS['mp']['fork'].Process):
        pass


    class ForkServerWorker(AbstractWorker, MP_CONTEXTS['mp']['forkserver'].Process):
        pass


class SpawnWorker(AbstractWorker, MP_CONTEXTS['mp']['spawn'].Process):
    pass


class ThreadingWorker(AbstractWorker, MP_CONTEXTS['threading'].Thread):
    pass


if DILL_INSTALLED:
    if FORK_AVAILABLE:
        class DillForkWorker(AbstractWorker, MP_CONTEXTS['mp']['fork'].Process):
            pass


        class DillForkServerWorker(AbstractWorker, MP_CONTEXTS['mp_dill']['forkserver'].Process):
            pass


    class DillSpawnWorker(AbstractWorker, MP_CONTEXTS['mp_dill']['spawn'].Process):
        pass


def worker_factory(start_method: str, use_dill: bool) -> Type[Union[AbstractWorker, mp.Process, Thread]]:
    """
    Returns the appropriate worker class given the start method

    :param start_method: What Process/Threading start method to use, see the WorkerPool constructor
    :param use_dill: Whether to use dill has serialization backend. Some exotic types (e.g., lambdas, nested functions)
        don't work well when using ``spawn`` as start method. In such cased, use ``dill`` (can be a bit slower
        sometimes)
    :return: Worker class
    """
    if start_method == 'threading':
        return ThreadingWorker
    elif use_dill:
        if not DILL_INSTALLED:
            raise ImportError("Can't use dill as the dependencies are not installed. Use `pip install mpire[dill]` to "
                              "install the required dependencies.")
        elif start_method == 'fork':
            if not FORK_AVAILABLE:
                raise ValueError("Start method 'fork' is not available")
            return DillForkWorker
        elif start_method == 'forkserver':
            if not FORK_AVAILABLE:
                raise ValueError("Start method 'forkserver' is not available")
            return DillForkServerWorker
        elif start_method == 'spawn':
            return DillSpawnWorker
        else:
            raise ValueError(f"Unknown start method with dill: '{start_method}'")
    else:
        if start_method == 'fork':
            if not FORK_AVAILABLE:
                raise ValueError("Start method 'fork' is not available")
            return ForkWorker
        elif start_method == 'forkserver':
            if not FORK_AVAILABLE:
                raise ValueError("Start method 'forkserver' is not available")
            return ForkServerWorker
        elif start_method == 'spawn':
            return SpawnWorker
        else:
            raise ValueError("Unknown start method: '{}'".format(start_method))



import _thread
import ctypes
import multiprocess as mp_dill
import multiprocessing as mp
import multiprocessing.synchronize
import queue
import threading
import unittest
import warnings
from datetime import datetime
from itertools import product
from unittest.mock import patch

from mpire.comms import NEW_MAP_PARAMS_PILL, NON_LETHAL_POISON_PILL, POISON_PILL, WorkerComms
from mpire.context import DEFAULT_START_METHOD, FORK_AVAILABLE, MP_CONTEXTS
from mpire.params import WorkerMapParams
from tests.utils import MockDatetimeNow


def _f1():
    pass


def _f2():
    pass


class WorkerCommsTest(unittest.TestCase):

    def test_init_comms(self):
        """
        Test if initializing/resetting the comms is done properly
        """
        test_ctx = [MP_CONTEXTS['mp_dill']['spawn'], MP_CONTEXTS['threading']]
        if FORK_AVAILABLE:
            test_ctx.extend([MP_CONTEXTS['mp_dill']['fork'], MP_CONTEXTS['mp']['forkserver']])

        for n_jobs, ctx in product([1, 2, 4], test_ctx):
            comms = WorkerComms(ctx, n_jobs)
            self.assertEqual(comms.ctx, ctx)
            self.assertEqual(comms.n_jobs, n_jobs)

            event_type = type(ctx.Event())
            lock_type = type(ctx.Lock())
            joinable_queue_type = type(ctx.JoinableQueue())

            with self.subTest('__init__ called', n_jobs=n_jobs):
                self.assertIsInstance(comms._keep_order, event_type)
                self.assertFalse(comms._keep_order.is_set())
                self.assertIsNone(comms._task_queues)
                self.assertIsNone(comms._task_idx)
                self.assertIsNone(comms._last_completed_task_worker_id)
                self.assertIsNone(comms._results_queue)
                self.assertListEqual(comms._exit_results_queues, [])
                self.assertIsNone(comms._all_exit_results_obtained)
                self.assertIsNone(comms._worker_done_array)
                self.assertIsNone(comms._workers_dead_locks)
                self.assertIsNone(comms._workers_dead)
                self.assertIsNone(comms._exception_queue)
                self.assertIsInstance(comms.exception_lock, lock_type)
                self.assertIsInstance(comms._exception_thrown, event_type)
                self.assertIsInstance(comms._kill_signal_received, event_type)
                self.assertFalse(comms._exception_thrown.is_set())
                self.assertFalse(comms._kill_signal_received.is_set())
                self.assertIsNone(comms._tasks_completed_array)
                self.assertIsNone(comms._tasks_completed_locks)
                self.assertIsNone(comms._progress_bar_last_updated)
                self.assertIsNone(comms._progress_bar_shutdown)
                self.assertIsNone(comms._progress_bar_complete)
                self.assertIsNone(comms._all_exit_results_obtained)

            with self.subTest('without initial values', n_jobs=n_jobs, has_worker_exit=False, has_progress_bar=False):
                comms.init_comms(has_worker_exit=False, has_progress_bar=False)
                self.assertFalse(comms._keep_order.is_set())
                self.assertEqual(len(comms._task_queues), n_jobs)
                for q in comms._task_queues:
                    self.assertIsInstance(q, joinable_queue_type)
                self.assertIsNone(comms._last_completed_task_worker_id)
                self.assertIsInstance(comms._results_queue, joinable_queue_type)
                self.assertListEqual(comms._exit_results_queues, [])
                self.assertIsInstance(comms._worker_done_array, ctypes.Array)
                self.assertEqual(len(comms._workers_dead), n_jobs)
                for worker_dead in comms._workers_dead:
                    self.assertIsInstance(worker_dead, event_type)
                    self.assertTrue(worker_dead.is_set())
                self.assertEqual(len(comms._workers_dead_locks), n_jobs)
                for worker_dead_lock in comms._workers_dead_locks:
                    self.assertIsInstance(worker_dead_lock, lock_type)
                self.assertIsInstance(comms._exception_queue, joinable_queue_type)
                self.assertFalse(comms._exception_thrown.is_set())
                self.assertFalse(comms._kill_signal_received.is_set())
                self.assertIsNone(comms._tasks_completed_array)
                self.assertIsNone(comms._tasks_completed_locks)
                self.assertIsNone(comms._progress_bar_last_updated)
                self.assertIsNone(comms._progress_bar_shutdown)
                self.assertIsNone(comms._progress_bar_complete)

                # Basic sanity checks for the values
                self.assertEqual(comms._task_idx, 0)
                self.assertEqual(list(comms._worker_done_array), [False for _ in range(n_jobs)])

            MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 0, 0, 0)]
            MockDatetimeNow.CURRENT_IDX = 0

            with self.subTest('without initial values', n_jobs=n_jobs, has_worker_exit=True, has_progress_bar=True), \
                    patch('mpire.comms.datetime', new=MockDatetimeNow):
                comms.init_comms(has_worker_exit=True, has_progress_bar=True)
                self.assertEqual(len(comms._exit_results_queues), n_jobs)
                for q in comms._exit_results_queues:
                    self.assertIsInstance(q, joinable_queue_type)
                self.assertIsInstance(comms._all_exit_results_obtained, event_type)
                self.assertFalse(comms._all_exit_results_obtained.is_set())
                self.assertIsInstance(comms._tasks_completed_array, ctypes.Array)
                self.assertEqual(len(comms._tasks_completed_locks), n_jobs)
                for lock in comms._tasks_completed_locks:
                    self.assertIsInstance(lock, lock_type)
                self.assertEqual(comms._progress_bar_last_updated, datetime(1970, 1, 1, 0, 0, 0, 0))
                self.assertIsInstance(comms._progress_bar_shutdown, event_type)
                self.assertFalse(comms._progress_bar_shutdown.is_set())
                self.assertIsInstance(comms._progress_bar_complete, event_type)
                self.assertFalse(comms._progress_bar_complete.is_set())

                # Basic sanity checks for the values
                self.assertEqual(list(comms._tasks_completed_array), [0 for _ in range(n_jobs)])

            # Set some values so we can test if the containers will be properly resetted
            comms._keep_order.set()
            comms._task_idx = 4
            comms._last_completed_task_worker_id = 2
            comms._worker_done_array[:] = [False, True, False, True][:n_jobs]
            [worker_dead.clear() for worker_dead in comms._workers_dead]
            comms._exception_thrown.set()
            comms._kill_signal_received.set()

            # Note that threading doesn't have a JoinableQueue, so they're taken from multiprocessing
            with self.subTest('with initial values', n_jobs=n_jobs, has_worker_exit=False, has_progress_bar=False):
                comms.init_comms(has_worker_exit=False, has_progress_bar=False)
                self.assertEqual(len(comms._task_queues), n_jobs)
                for q in comms._task_queues:
                    self.assertIsInstance(q, joinable_queue_type)
                self.assertIsNone(comms._last_completed_task_worker_id)
                self.assertIsInstance(comms._results_queue, joinable_queue_type)
                self.assertListEqual(comms._exit_results_queues, [])
                self.assertIsInstance(comms._worker_done_array, ctypes.Array)
                self.assertEqual(len(comms._workers_dead), n_jobs)
                for worker_dead in comms._workers_dead:
                    self.assertIsInstance(worker_dead, event_type)
                    self.assertTrue(worker_dead.is_set())
                self.assertEqual(len(comms._workers_dead_locks), n_jobs)
                for worker_dead_lock in comms._workers_dead_locks:
                    self.assertIsInstance(worker_dead_lock, lock_type)
                self.assertIsInstance(comms._exception_queue, joinable_queue_type)
                self.assertFalse(comms._exception_thrown.is_set())
                self.assertFalse(comms._kill_signal_received.is_set())
                self.assertIsNone(comms._tasks_completed_array)
                self.assertIsNone(comms._tasks_completed_locks)
                self.assertIsNone(comms._progress_bar_last_updated)
                self.assertIsNone(comms._progress_bar_shutdown)
                self.assertIsNone(comms._progress_bar_complete)

                # Some variables are not reset by this function, but are reset otherwise
                self.assertIsInstance(comms._keep_order, event_type)
                self.assertTrue(comms._keep_order.is_set())

                # Basic sanity checks for the values
                self.assertEqual(comms._task_idx, 0)
                self.assertEqual(list(comms._worker_done_array), [False for _ in range(n_jobs)])

    def test_progress_bar(self):
        """
        Test progress bar related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2)

        # Has progress bar
        self.assertFalse(comms.has_progress_bar())
        comms.init_comms(False, True)
        self.assertTrue(comms.has_progress_bar())

        # Nothing available yet
        self.assertEqual(sum(comms._tasks_completed_array), 0)

        MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 0, 0)]
        MockDatetimeNow.CURRENT_IDX = 0

        # 3 task done, but not enough time has passed to send the update
        last_updated = datetime(1970, 1, 1, 0, 0, 0, 0)
        n_tasks_completed = 0
        with patch('mpire.comms.datetime', new=MockDatetimeNow):
            for n in range(1, 4):
                last_updated, n_tasks_completed = comms.task_completed_progress_bar(0, last_updated, n_tasks_completed,
                                                                                    force_update=False)
                self.assertEqual(n_tasks_completed, n)
        self.assertEqual(sum(comms._tasks_completed_array), 0)

        # Not enough time has passed, but we'll force the update. Number of tasks done should still be 3
        with patch('mpire.comms.datetime', new=MockDatetimeNow):
            last_updated, n_tasks_completed = comms.task_completed_progress_bar(0, last_updated, n_tasks_completed,
                                                                                force_update=True)
        self.assertEqual(comms.get_tasks_completed_progress_bar(), 3)
        self.assertEqual(last_updated, datetime(1970, 1, 1, 0, 0, 0, 0))
        self.assertEqual(n_tasks_completed, 0)

        # 4 tasks already done and another 4 tasks done. Enough time should've passed for each update call, except the
        # second. In total we have 3 (from above) + 4 + 4 = 11 tasks done
        MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 1, 0, 0),
                                         datetime(1970, 1, 1, 0, 1, 0, 0),
                                         datetime(1970, 1, 1, 0, 3, 0, 0),
                                         datetime(1970, 1, 1, 0, 4, 0, 0)]
        MockDatetimeNow.CURRENT_IDX = 0
        last_updated = datetime(1970, 1, 1, 0, 0, 0, 0)
        n_tasks_completed = 4
        with patch('mpire.comms.datetime', new=MockDatetimeNow):
            for _ in range(4):
                last_updated, n_tasks_completed = comms.task_completed_progress_bar(0, last_updated, n_tasks_completed,
                                                                                    force_update=False)
                self.assertEqual(last_updated, MockDatetimeNow.RETURN_VALUES[MockDatetimeNow.CURRENT_IDX - 1])
        self.assertEqual(comms.get_tasks_completed_progress_bar(), 11)
        self.assertEqual(last_updated, datetime(1970, 1, 1, 0, 4, 0, 0))
        self.assertEqual(n_tasks_completed, 0)

        # Signal shutdown
        comms.signal_progress_bar_shutdown()
        self.assertEqual(comms.get_tasks_completed_progress_bar(), POISON_PILL)
        comms._progress_bar_shutdown.clear()

        # Set exception
        comms.signal_exception_thrown()
        self.assertEqual(comms.get_tasks_completed_progress_bar(), POISON_PILL)
        comms._exception_thrown.clear()

        # Set kill signal received
        comms.signal_kill_signal_received()
        self.assertEqual(comms.get_tasks_completed_progress_bar(), POISON_PILL)
        comms._kill_signal_received.clear()

    def test_progress_bar_shutdown(self):
        """
        Test progress bar complete related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2)
        comms.init_comms(False, True)

        self.assertFalse(comms._progress_bar_shutdown.is_set())
        comms.signal_progress_bar_shutdown()
        self.assertTrue(comms._progress_bar_shutdown.is_set())

    def test_progress_bar_complete(self):
        """
        Test progress bar complete related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2)
        comms.init_comms(False, True)

        self.assertFalse(comms._progress_bar_complete.is_set())
        comms.signal_progress_bar_complete()
        self.assertTrue(comms._progress_bar_complete.is_set())

        with patch.object(comms._progress_bar_complete, 'wait') as p:
            comms.wait_until_progress_bar_is_complete()
            self.assertEqual(p.call_count, 1)

    def test_keep_order(self):
        """
        Test keep_order related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2)

        self.assertFalse(comms.keep_order())
        comms.signal_keep_order()
        self.assertTrue(comms.keep_order())
        comms.clear_keep_order()
        self.assertFalse(comms.keep_order())

    def test_tasks(self):
        """
        Test task related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 3)
        comms.init_comms(False, False)

        # Nothing available yet
        for worker_id in range(3):
            with self.assertRaises(queue.Empty):
                comms._task_queues[worker_id].get(block=False)

        # Add a few tasks. As no worker has completed tasks yet, it should give the task to workers in order
        comms.add_task(12)
        comms.add_task('hello world')
        comms.add_task({'foo': 'bar'})
        comms.add_task({'foo': 'baz'})
        comms.add_task(34.43)
        comms.add_task(datetime(2000, 1, 1, 1, 2, 3))
        tasks = []
        for worker_id in [0, 1, 2, 0, 1, 2]:
            tasks.append(comms.get_task(worker_id))
            comms.task_done(worker_id)
        self.assertListEqual(tasks, [12, 'hello world', {'foo': 'bar'}, {'foo': 'baz'}, 34.43,
                                     datetime(2000, 1, 1, 1, 2, 3)])

        # When workers have completed tasks, it depends on the last one who gets the new task. After giving a task to
        # that worker the worker ID that last completed a task is reset again. So, it should continue with the normal
        # order
        comms.init_comms(False, False)
        comms.add_task(12)
        comms.add_task('hello world')
        comms.add_task({'foo': 'bar'})
        comms._last_completed_task_worker_id = 2
        comms.add_task({'foo': 'baz'})
        comms._last_completed_task_worker_id = 1
        comms.add_task(34.43)
        comms._last_completed_task_worker_id = 0
        comms.add_task(datetime(2000, 1, 1, 1, 2, 3))
        comms._last_completed_task_worker_id = 2
        comms.add_task('123')
        comms.add_task(123)
        comms.add_task(1337)
        tasks = []
        for worker_id in [0, 1, 2, 2, 1, 0, 2, 0, 1]:
            tasks.append(comms.get_task(worker_id))
            comms.task_done(worker_id)
        self.assertListEqual(tasks, [12, 'hello world', {'foo': 'bar'}, {'foo': 'baz'}, 34.43,
                                     datetime(2000, 1, 1, 1, 2, 3), '123', 123, 1337])

        # Throw in an exception. Should return None
        comms.signal_exception_thrown()
        for worker_id in range(3):
            self.assertIsNone(comms.get_task(worker_id))

        # Should be joinable
        comms.join_task_queues()

    def test_results(self):
        """
        Test results related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2)
        comms.init_comms(False, False)

        # Nothing available yet
        with self.assertRaises(queue.Empty):
            comms._results_queue.get(block=False)

        # Add a few results. Note that `get_results` calls `task_done`
        comms.add_results(0, 12)
        comms.add_results(1, 'hello world')
        comms.add_results(1, {'foo': 'bar'})
        comms.add_results(0, '123')
        self.assertEqual(comms.get_results(), 12)
        self.assertEqual(comms._last_completed_task_worker_id, 0)
        self.assertEqual(comms.get_results(), 'hello world')
        self.assertEqual(comms._last_completed_task_worker_id, 1)
        self.assertEqual(comms.get_results(), {'foo': 'bar'})
        self.assertEqual(comms._last_completed_task_worker_id, 1)
        self.assertEqual(comms.get_results(), '123')
        self.assertEqual(comms._last_completed_task_worker_id, 0)

        # Should be joinable. When using keep_alive, it should still be open
        comms.join_results_queues(keep_alive=True)
        comms.add_results(0, 12)
        comms.get_results()
        comms.join_results_queues(keep_alive=False)

        # Depending on Python version it either throws AssertionError or ValueError
        with self.assertRaises((AssertionError, ValueError)):
            comms.add_results(0, 12)

    def test_exit_results(self):
        """
        Test exit results related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 3)
        comms.init_comms(True, False)

        # Nothing available yet
        for worker_id in range(3):
            with self.assertRaises(queue.Empty):
                comms.get_exit_results(worker_id, block=False)

        # Add a few results. Note that `get_exit_results` calls `task_done`
        for worker_id in range(3):
            comms.add_exit_results(worker_id, worker_id)
            comms.add_exit_results(worker_id, 'hello world')
            comms.add_exit_results(worker_id, {'foo': 'bar'})
        self.assertListEqual([comms.get_exit_results(worker_id=0) for _ in range(3)],
                             [0, 'hello world', {'foo': 'bar'}])
        self.assertListEqual([comms.get_exit_results(worker_id=1) for _ in range(3)],
                             [1, 'hello world', {'foo': 'bar'}])
        self.assertListEqual([comms.get_exit_results(worker_id=2) for _ in range(3)],
                             [2, 'hello world', {'foo': 'bar'}])

        # When an exception is thrown it should return directly. However, when block=False it should still get it. We
        # use a while loop because we need to wait until the queue is ready. This block=False trick is used in
        # terminate().
        comms.add_exit_results(0, 'hello world')
        comms.signal_exception_thrown()
        self.assertIsNone(comms.get_exit_results(0))
        while True:
            try:
                self.assertEqual(comms.get_exit_results(0, block=False), 'hello world')
                break
            except queue.Empty:
                pass
        comms._exception_thrown.clear()

        # Should be joinable. When using keep_alive, it should still be open
        comms.join_results_queues(keep_alive=True)
        for worker_id in range(3):
            comms.add_exit_results(worker_id, 'hello world')
            comms.get_exit_results(worker_id)
        comms.join_results_queues(keep_alive=False)

        # Depending on Python version it either throws AssertionError or ValueError
        for worker_id in range(3):
            with self.assertRaises((AssertionError, ValueError)):
                comms.add_exit_results(worker_id, 'hello world')

    def test_exit_results_all_workers(self):
        """
        Test exit results related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 4)
        comms.init_comms(True, False)

        # Add a few results. Every worker will always have a return value (even if it's the implicit None). Note that
        # `get_exit_results` calls `task_done`
        for worker_id in range(3):
            comms.add_exit_results(worker_id, worker_id)
        comms.add_exit_results(3, None)
        self.assertListEqual(comms.get_exit_results_all_workers(), [0, 1, 2, None])

        # Should be joinable
        comms.join_results_queues()

    def test_exit_results_all_workers_exception_thrown(self):
        """
        Test exit results related functions. When an exception occurred, it should return an empty list
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 3)
        comms.init_comms(True, False)

        # Add a few results.
        for worker_id in range(3):
            comms.add_exit_results(worker_id, worker_id)

        # Set exception
        comms.signal_exception_thrown()

        # Should return empty list
        self.assertListEqual(comms.get_exit_results_all_workers(), [])

        # Drain and join
        comms._exception_thrown.clear()
        comms.get_exit_results_all_workers()
        comms.join_results_queues()

    def test_all_exit_results_obtained(self):
        """
        Test all exit results obtained related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2)
        comms.init_comms(True, False)

        self.assertFalse(comms._all_exit_results_obtained.is_set())
        comms.signal_all_exit_results_obtained()
        self.assertTrue(comms._all_exit_results_obtained.is_set())

        with patch.object(comms._all_exit_results_obtained, 'wait') as p:
            comms.wait_until_all_exit_results_obtained()
            self.assertEqual(p.call_count, 1)

    def test_add_new_map_params(self):
        """
        Test new map parameters function
        """
        comms = WorkerComms(MP_CONTEXTS['mp_dill'][DEFAULT_START_METHOD], 2)
        comms.init_comms(False, False)

        map_params = WorkerMapParams(_f1, None, None, 1)
        comms.add_new_map_params(map_params)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for worker_id in range(2):
                self.assertEqual(comms.get_task(worker_id), NEW_MAP_PARAMS_PILL)
                self.assertEqual(comms.get_task(worker_id), map_params)
                comms.task_done(worker_id)
                comms.task_done(worker_id)

        map_params = WorkerMapParams(_f1, _f2, None, None)
        comms.add_new_map_params(map_params)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for worker_id in range(2):
                self.assertEqual(comms.get_task(worker_id), NEW_MAP_PARAMS_PILL)
                self.assertEqual(comms.get_task(worker_id), map_params)
                comms.task_done(worker_id)
                comms.task_done(worker_id)

    def test_exceptions(self):
        """
        Test exception related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2)
        comms.init_comms(False, False)

        # Nothing available yet
        with self.assertRaises(queue.Empty):
            comms._exception_queue.get(block=False)

        # Add a few exceptions
        comms.add_exception(TypeError, 'TypeError')
        comms.add_exception(ValueError, 'ValueError')
        comms.add_exception(RuntimeError, 'RuntimeError')
        self.assertListEqual([comms.get_exception() for _ in range(3)],
                             [(TypeError, 'TypeError'), (ValueError, 'ValueError'), (RuntimeError, 'RuntimeError')])
        [comms.task_done_exception() for _ in range(3)]

        # Add poison pill
        comms.add_exception_poison_pill()
        self.assertEqual(comms.get_exception(), (POISON_PILL, POISON_PILL))
        comms.task_done_exception()

        # Should be joinable. When using keep_alive, it should still be open
        comms.join_exception_queue(keep_alive=True)
        comms.add_exception(TypeError, 'TypeError')
        comms.get_exception()
        comms.task_done_exception()
        comms.join_exception_queue(keep_alive=False)

        # Depending on Python version it either throws AssertionError or ValueError
        with self.assertRaises((AssertionError, ValueError)):
            comms.add_exception(TypeError, 'TypeError')

    def test_exception_thrown(self):
        """
        Test exception thrown related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2)

        self.assertFalse(comms.exception_thrown())
        comms.signal_exception_thrown()
        self.assertTrue(comms.exception_thrown())
        comms._exception_thrown.clear()
        self.assertFalse(comms.exception_thrown())

    def test_kill_signal_received(self):
        """
        Test kill signal received related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2)

        self.assertFalse(comms.kill_signal_received())
        comms.signal_kill_signal_received()
        self.assertTrue(comms.kill_signal_received())
        comms._kill_signal_received.clear()
        self.assertFalse(comms.kill_signal_received())

    def test_worker_poison_pill(self):
        """
        Test that a poison pill is inserted for every worker
        """
        for n_jobs in [1, 2, 4]:
            with self.subTest(n_jobs=n_jobs):
                comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], n_jobs)
                comms.init_comms(False, False)
                comms.insert_poison_pill()
                for worker_id in range(n_jobs):
                    self.assertEqual(comms.get_task(worker_id), POISON_PILL)
                    comms.task_done(worker_id)
                comms.join_task_queues()

    def test_worker_non_lethal_poison_pill(self):
        """
        Test that a non-lethal poison pill is inserted for every worker
        """
        # Shouldn't be the same thing
        self.assertNotEqual(POISON_PILL, NON_LETHAL_POISON_PILL)

        for n_jobs in [1, 2, 4]:
            with self.subTest(n_jobs=n_jobs):
                comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], n_jobs)
                comms.init_comms(False, False)
                comms.insert_non_lethal_poison_pill()
                for worker_id in range(n_jobs):
                    self.assertEqual(comms.get_task(worker_id), NON_LETHAL_POISON_PILL)
                    comms.task_done(worker_id)
                comms.join_task_queues()

    def test_worker_restart(self):
        """
        Test worker restart related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 5)
        comms.init_comms(False, False)

        # No restarts yet
        self.assertListEqual(list(comms.get_worker_restarts()), [])

        # Signal some restarts
        comms.signal_worker_restart(0)
        comms.signal_worker_restart(2)
        comms.signal_worker_restart(3)

        # Restarts available
        self.assertListEqual(list(comms.get_worker_restarts()), [0, 2, 3])

        # Reset some
        comms.reset_worker_restart(0)
        comms.reset_worker_restart(3)

        # Restarts available
        self.assertListEqual(list(comms.get_worker_restarts()), [2])

        # Reset last one
        comms.reset_worker_restart(2)
        self.assertListEqual(list(comms.get_worker_restarts()), [])

    def test_worker_alive(self):
        """
        Test worker alive related functions
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 5)
        comms.init_comms(False, False)

        # Signal some workers are alive
        comms.signal_worker_alive(0)
        comms.signal_worker_alive(1)
        comms.signal_worker_dead(1)
        comms.signal_worker_alive(2)
        comms.signal_worker_alive(3)

        # Check alive status
        self.assertListEqual([comms.is_worker_alive(worker_id) for worker_id in range(5)],
                             [True, False, True, True, False])

        # Reset some
        comms.signal_worker_dead(0)
        comms.signal_worker_dead(3)

        # Check alive status
        self.assertListEqual([comms.is_worker_alive(worker_id) for worker_id in range(5)],
                             [False, False, True, False, False])

        # We test wait by simply checking the call count
        for worker_id in range(5):
            with patch.object(comms._workers_dead[worker_id], 'wait') as p:
                comms.wait_for_dead_worker(worker_id)
                self.assertEqual(p.call_count, 1)

    def test_drain_queues_terminate_worker(self):
        """
        get_results should be called once, get_exit_results should be called when exit function is defined
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 5)
        dont_wait_event = threading.Event()
        dont_wait_event.set()

        comms.init_comms(has_worker_exit=False, has_progress_bar=False)
        with self.subTest(has_worker_exit=False, has_progress_bar=False), \
                patch.object(comms, 'get_results', side_effect=comms.get_results) as p1, \
                patch.object(comms, 'get_exit_results') as p2:
            comms.drain_queues_terminate_worker(0, dont_wait_event)
            self.assertEqual(p1.call_count, 1)
            self.assertEqual(p2.call_count, 0)
            self.assertTrue(dont_wait_event.is_set())

        comms.init_comms(has_worker_exit=True, has_progress_bar=False)
        for worker_id in range(5):
            with self.subTest(has_worker_exit=True, has_progress_bar=False, worker_id=worker_id), \
                    patch.object(comms, 'get_results', side_effect=comms.get_results) as p1, \
                    patch.object(comms, 'get_exit_results', side_effect=comms.get_exit_results) as p2:
                comms.drain_queues_terminate_worker(worker_id, dont_wait_event)
                self.assertEqual(p1.call_count, 1)
                self.assertEqual(p2.call_count, 1)
                self.assertEqual(p2.call_args_list[0][0][0], worker_id)
                self.assertTrue(dont_wait_event.is_set())

        comms.init_comms(has_worker_exit=False, has_progress_bar=True)
        with self.subTest(has_worker_exit=False, has_progress_bar=True), \
                patch.object(comms, 'get_results', side_effect=comms.get_results) as p1, \
                patch.object(comms, 'get_exit_results') as p2:
            comms.drain_queues_terminate_worker(0, dont_wait_event)
            self.assertEqual(p1.call_count, 1)
            self.assertEqual(p2.call_count, 0)
            self.assertTrue(dont_wait_event.is_set())

        comms.init_comms(has_worker_exit=True, has_progress_bar=True)
        for worker_id in range(5):
            with self.subTest(has_worker_exit=True, has_progress_bar=True, worker_id=worker_id), \
                    patch.object(comms, 'get_results', side_effect=comms.get_results) as p1, \
                    patch.object(comms, 'get_exit_results', side_effect=comms.get_exit_results) as p2:
                comms.drain_queues_terminate_worker(worker_id, dont_wait_event)
                self.assertEqual(p1.call_count, 1)
                self.assertEqual(p2.call_count, 1)
                self.assertEqual(p2.call_args_list[0][0][0], worker_id)
                self.assertTrue(dont_wait_event.is_set())

    def test_drain_queues(self):
        """
        _drain_and_join_queue should be called for every queue that matters. There are as many tasks queues as workers,
        1 results queue, if a worker_exit function has been provided as many exit results queues as workers, and if a
        progress bar has been enabled one queu for the progress bar.
        """
        for n_jobs, has_worker_exit, has_progress_bar in product([1, 2, 4], [False, True], [False, True]):
            comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], n_jobs)
            comms.init_comms(has_worker_exit=has_worker_exit, has_progress_bar=has_progress_bar)

            with self.subTest(n_jobs=n_jobs, has_worker_exit=has_worker_exit, has_progress_bar=has_progress_bar), \
                    patch.object(comms, 'drain_and_join_queue') as p:
                comms.drain_queues()
                self.assertEqual(p.call_count,
                                 n_jobs + 1 + (n_jobs if has_worker_exit else 0))

    def test__drain_and_join_queue(self):
        """
        Test draining queues
        """
        comms = WorkerComms(MP_CONTEXTS['mp'][DEFAULT_START_METHOD], 2)

        # Create a custom queue with some data
        q = mp.JoinableQueue()
        q.put(1)
        q.put('hello')
        q.put('world')

        # Drain queue. It should now be closed
        comms._drain_and_join_queue(q)

        # Depending on Python version it either throws OSError or ValueError
        with self.assertRaises((OSError, ValueError)):
            q.get(block=False)



import logging
from ctypes import c_char
from multiprocessing import Array, Event, Lock
from multiprocessing.managers import SyncManager

from typing import Optional, Tuple

from mpire.signal import DisableKeyboardInterruptSignal

TqdmConnectionDetails = Tuple[Optional[bytes], bool]

logger = logging.getLogger(__name__)


class TqdmPositionRegister:

    """
    Class that keeps track of all the registered progress bar positions. Needed to properly display multiple tqdm
    progress bars
    """

    def __init__(self) -> None:
        self.lock = Lock()
        self.highest_position = None

    def register_progress_bar_position(self, position) -> bool:
        """
        Register new progress bar position. Returns True when it's the first one to register

        :param position: Progress bar position
        :return: Whether this progress bar is the first one to register
        """
        with self.lock:
            first_one = self.highest_position is None
            if self.highest_position is None or position > self.highest_position:
                self.highest_position = position

        return first_one

    def get_highest_progress_bar_position(self) -> Optional[int]:
        """
        Obtain the highest registered progress bar position

        :return: Highest progress bar position
        """
        with self.lock:
            return self.highest_position

    def reset_progress_bar_positions(self) -> None:
        """
        Reset the registered progress bar positions
        """
        with self.lock:
            self.highest_position = None


class TqdmLock:

    """
    Small wrapper around a multiprocessing.Lock proxy object returned by a Manager. These proxies only expose public
    functions, so we wrap such a proxy and expose the private/public functions needed for a lock to work.
    """

    def __init__(self, lock_proxy: Lock) -> None:
        self.lock_proxy = lock_proxy

    def acquire(self, *args, **kwargs) -> None:
        self.lock_proxy.acquire(*args, **kwargs)

    def release(self) -> None:
        self.lock_proxy.release()

    def __enter__(self) -> None:
        self.lock_proxy.acquire()

    def __exit__(self, *_) -> None:
        self.lock_proxy.release()


class TqdmManager:

    LOCK = Lock()
    POSITION_REGISTER = TqdmPositionRegister()

    MANAGER = None
    MANAGER_HOST = Array(c_char, 10000, lock=True)
    MANAGER_STARTED = Event()

    def __init__(self) -> None:
        """
        Tqdm manager wrapper for syncing multiple progress bars, independent of process start method used.
        """
        # Connect to existing manager, if it exists
        if self.MANAGER_STARTED.is_set():
            self.connect_to_manager()

    @classmethod
    def start_manager(cls) -> bool:
        """
        Sets up and starts the tqdm manager

        :return: Whether the manager was started
        """
        # Don't do anything when there's already a connected tqdm manager
        if cls.MANAGER_STARTED.is_set():
            return False

        logger.debug("Starting TQDM manager")

        # Create manager
        with DisableKeyboardInterruptSignal():
            cls.MANAGER = SyncManager(authkey=b'mpire_tqdm')
            cls.MANAGER.register('get_tqdm_lock', cls._get_tqdm_lock)
            cls.MANAGER.register('get_tqdm_position_register', cls._get_tqdm_position_register)
            cls.MANAGER.start()
        cls.MANAGER_STARTED.set()

        # Set host so other processes know where to connect to. On some systems and Python versions address is a bytes
        # object, on others it's a string. On some it's also prefixed by a null byte which needs to be removed (null
        # byte doesn't work with Array).
        address = cls.MANAGER.address
        if isinstance(address, str):
            address = address.encode()
        if address[0] == 0:
            address = address[1:]
        cls.MANAGER_HOST.value = address

        return True

    def connect_to_manager(self) -> None:
        """
        Connect to the tqdm manager
        """
        # Connect to a server. On some systems and Python versions the address is prefixed by a null byte (which was
        # stripped when setting the host value, due to restrictions in Array). Address needs to be a string.
        address = self.MANAGER_HOST.value.decode()
        try:
            self.MANAGER = SyncManager(address=address, authkey=b'mpire_tqdm')
            self.MANAGER.register('get_tqdm_lock')
            self.MANAGER.register('get_tqdm_position_register')
            self.MANAGER.connect()
        except FileNotFoundError:
            address = f"\x00{address}"
            self.MANAGER = SyncManager(address=address, authkey=b'mpire_tqdm')
            self.MANAGER.register('get_tqdm_lock')
            self.MANAGER.register('get_tqdm_position_register')
            self.MANAGER.connect()

    @classmethod
    def stop_manager(cls) -> None:
        """
        Stops the tqdm manager
        """
        cls.MANAGER.shutdown()
        cls.MANAGER = None
        cls.MANAGER_HOST.value = b''
        cls.MANAGER_STARTED.clear()

    @staticmethod
    def _get_tqdm_lock() -> Lock:
        """
        This function needs to be static, because a manager doesn't pass arguments to registered functions

        :return: Lock object for tqdm
        """
        return TqdmManager.LOCK

    @staticmethod
    def _get_tqdm_position_register() -> TqdmPositionRegister:
        """
        This function needs to be static, because a manager doesn't pass arguments to registered functions

        :return: tqdm position register
        """
        return TqdmManager.POSITION_REGISTER

    def get_lock_and_position_register(self) -> Tuple[TqdmLock, TqdmPositionRegister]:
        """
        Obtain synchronized tqdm lock and positions register

        :return: Synchronized tqdm lock and tqdm position register
        """
        return TqdmLock(self.MANAGER.get_tqdm_lock()), self.MANAGER.get_tqdm_position_register()

    @classmethod
    def get_connection_details(cls) -> TqdmConnectionDetails:
        """
        Obtains the connection details of the tqdm manager. These details are needed to be passed on to child process
        when the start method is either forkserver or spawn.

        :return: TQDM manager host and whether a manager is started/connected
        """
        return cls.MANAGER_HOST.value, cls.MANAGER_STARTED.is_set()

    @classmethod
    def set_connection_details(cls, tqdm_connection_details: TqdmConnectionDetails) -> None:
        """
        Sets the tqdm connection details and connects to an existing tqdm manager if needed.

        :param tqdm_connection_details: TQDM manager host, and whether a manager is started/connected
        """
        tqdm_manager_host, tqdm_manager_started = tqdm_connection_details
        if not cls.MANAGER_STARTED.is_set():
            cls.MANAGER_HOST.value = tqdm_manager_host
            if tqdm_manager_started:
                cls.MANAGER_STARTED.set()



import logging
import multiprocessing as mp
import queue
import threading
import time
from datetime import datetime
from typing import Any, Generator, List, Optional, Tuple, Union

from mpire.params import WorkerMapParams
from mpire.signal import DelayedKeyboardInterrupt

logger = logging.getLogger(__name__)

POISON_PILL = '\0'
NON_LETHAL_POISON_PILL = '\1'
NEW_MAP_PARAMS_PILL = '\2'


class WorkerComms:

    """
    Class that contains all the inter-process communication objects (locks, events, queues, etc.) and functionality to
    interact with them, except for the worker insights comms.

    Contains:
    - Progress bar comms
    - Tasks & (exit) results comms
    - Exception handling comms
    - Terminating and restarting comms
    """

    # Amount of time in between each progress bar update
    progress_bar_update_interval = 0.1

    def __init__(self, ctx: mp.context.BaseContext, n_jobs: int) -> None:
        """
        :param ctx: Multiprocessing context
        :param n_jobs: Number of workers
        """
        self.ctx = ctx
        self.n_jobs = n_jobs

        # Whether or not to inform the child processes to keep order in mind (for the map functions)
        self._keep_order = self.ctx.Event()

        # Queue to pass on tasks to child processes
        self._task_queues = None
        self._task_idx = None
        self._last_completed_task_worker_id = None

        # Queue where the child processes can pass on results
        self._results_queue = None

        # Queue where the child processes can store exit results in
        self._exit_results_queues = []
        self._all_exit_results_obtained = None

        # Array where the child processes can request a restart
        self._worker_done_array = None

        # List of Event objects to indicate whether workers are alive, together with accompanying locks
        self._workers_dead = None
        self._workers_dead_locks = None

        # Queue where the child processes can pass on an encountered exception
        self._exception_queue = None

        # Lock object such that child processes can only throw one at a time. The Event object ensures only one
        # exception can be thrown. When the threading backend is used we switch to a multiprocessing Event, because the
        # progress bar handler needs a process-aware object
        self.exception_lock = self.ctx.Lock()
        self._exception_thrown = self.ctx.Event()
        self._kill_signal_received = self.ctx.Event()

        # Array where the number of completed tasks is stored for the progress bar. We don't use the vanilla lock from
        # multiprocessing.Array, but create a lock per worker such that workers can write concurrently.
        self._tasks_completed_array = None
        self._tasks_completed_locks = None
        self._progress_bar_last_updated = None
        self._progress_bar_shutdown = None
        self._progress_bar_complete = None

    ################
    # Initialization
    ################

    def init_comms(self, has_worker_exit: bool, has_progress_bar: bool) -> None:
        """
        Initialize/Reset comms containers.

        Threading doesn't have a JoinableQueue, so the threading context returns a multiprocessing.JoinableQueue
        instead. However, in the (unlikely) scenario that threading does get one, we explicitly switch to a
        multiprocessing.JoinableQueue for both the exception queue and progress bar tasks completed queue, because the
        progress bar handler needs process-aware objects.

        :param has_worker_exit: Whether there's a worker_exit function provided
        :param has_progress_bar: Whether there's a progress bar
        """
        # Task related
        self._task_queues = [self.ctx.JoinableQueue() for _ in range(self.n_jobs)]
        self.reset_last_completed_task_info()

        # Results related
        self._results_queue = self.ctx.JoinableQueue()
        if has_worker_exit:
            self._exit_results_queues = [self.ctx.JoinableQueue() for _ in range(self.n_jobs)]
            self._all_exit_results_obtained = self.ctx.Event()
        else:
            self._exit_results_queues = []
            self._all_exit_results_obtained = None

        # Worker status
        self._worker_done_array = self.ctx.Array('b', self.n_jobs, lock=False)
        self._workers_dead = [self.ctx.Event() for _ in range(self.n_jobs)]
        [worker_dead.set() for worker_dead in self._workers_dead]
        self._workers_dead_locks = [self.ctx.Lock() for _ in range(self.n_jobs)]

        # Exception related
        self._exception_queue = self.ctx.JoinableQueue()
        self._exception_thrown.clear()
        self._kill_signal_received.clear()

        # Progress bar related
        if has_progress_bar:
            self._tasks_completed_array = self.ctx.Array('L', self.n_jobs, lock=False)  # ULong (should be enough)
            self._tasks_completed_locks = [self.ctx.Lock() for _ in range(self.n_jobs)]
            self._progress_bar_last_updated = datetime.now()
            self._progress_bar_shutdown = self.ctx.Event()
            self._progress_bar_complete = self.ctx.Event()
        else:
            self._tasks_completed_array = None
            self._tasks_completed_locks = None
            self._progress_bar_last_updated = None
            self._progress_bar_shutdown = None
            self._progress_bar_complete = None

    def reset_last_completed_task_info(self) -> None:
        """
        Resets the task_idx and last_completed_task_worker_id
        """
        self._task_idx = 0
        self._last_completed_task_worker_id = None

    ################
    # Progress bar
    ################

    def has_progress_bar(self) -> bool:
        """
        :return: Whether we have a progress bar
        """
        return self._tasks_completed_array is not None

    def task_completed_progress_bar(self, worker_id: int, progress_bar_last_updated: datetime,
                                    progress_bar_n_tasks_completed: Optional[int] = None,
                                    force_update: bool = False) -> Tuple[datetime, int]:
        """
        Signal that we've completed a task every 0.1 seconds, for the progress bar

        :param worker_id: Worker ID
        :param progress_bar_last_updated: Last time the progress bar update was send
        :param progress_bar_n_tasks_completed: Number of tasks completed since last update
        :param force_update: Whether to force an update
        :return: Tuple containing new last updated time and number of tasks completed since last update
        """
        # If it's not forced we updated the number of completed tasks
        if not force_update:
            progress_bar_n_tasks_completed += 1

        # Check if we need to update
        now = datetime.now()
        if force_update or (now - progress_bar_last_updated).total_seconds() > self.progress_bar_update_interval:
            with self._tasks_completed_locks[worker_id]:
                self._tasks_completed_array[worker_id] += progress_bar_n_tasks_completed
            progress_bar_last_updated = now
            progress_bar_n_tasks_completed = 0

        return progress_bar_last_updated, progress_bar_n_tasks_completed

    def get_tasks_completed_progress_bar(self) -> Union[int, str]:
        """
        Obtain the number of tasks completed by the workers. As the progress bar handler lives inside a thread we don't
        poll continuously, but every 0.1 seconds.

        :return: The number of tasks done or a poison pill
        """
        # Check if we need to wait a bit for the next update
        time_diff = (datetime.now() - self._progress_bar_last_updated).total_seconds()
        if time_diff < self.progress_bar_update_interval:
            time.sleep(self.progress_bar_update_interval - time_diff)

        # Sum the tasks completed and return
        while (not self.exception_thrown() and not self.kill_signal_received() and
               not self._progress_bar_shutdown.is_set()):
            n_tasks_completed = 0
            for worker_id in range(self.n_jobs):
                with self._tasks_completed_locks[worker_id]:
                    n_tasks_completed += self._tasks_completed_array[worker_id]
            self._progress_bar_last_updated = datetime.now()
            return n_tasks_completed

        return POISON_PILL

    def signal_progress_bar_shutdown(self) -> None:
        """
        Signals the progress bar handling process to shut down
        """
        self._progress_bar_shutdown.set()

    def signal_progress_bar_complete(self) -> None:
        """
        Signal that the progress bar is complete
        """
        self._progress_bar_complete.set()

    def wait_until_progress_bar_is_complete(self) -> None:
        """
        Waits until the progress bar is completed
        """
        while not self.exception_thrown():
            if self._progress_bar_complete.wait(timeout=0.01):
                return

    ################
    # Order modifiers
    ################

    def signal_keep_order(self) -> None:
        """
        Set that we need to keep order in mind
        """
        return self._keep_order.set()

    def clear_keep_order(self) -> None:
        """
        Forget that we need to keep order in mind
        """
        return self._keep_order.clear()

    def keep_order(self) -> bool:
        """
        :return: Whether we need to keep order in mind
        """
        return self._keep_order.is_set()

    ################
    # Tasks & results
    ################

    def add_task(self, task: Any, worker_id: Optional[int] = None) -> None:
        """
        Add a task to the queue so a worker can process it.

        :param task: A tuple of arguments to pass to a worker, which acts upon it
        :param worker_id: If provided, give the task to the worker ID
        """
        # When a worker ID is not present, we check whether we got results already. If so, we give the next task to the
        # worker who completed that task. Otherwise, we decide based on order
        if worker_id is None:
            if self._last_completed_task_worker_id is not None:
                worker_id = self._last_completed_task_worker_id
                self._last_completed_task_worker_id = None
            else:
                worker_id = self._task_idx % self.n_jobs
                self._task_idx += 1

        with DelayedKeyboardInterrupt():
            self._task_queues[worker_id].put(task, block=True)

    def get_task(self, worker_id: int) -> Any:
        """
        Obtain new chunk of tasks. Occasionally we check if an exception has been thrown. If so, we should quit.

        :param worker_id: Worker ID
        :return: Chunk of tasks or None when an exception was thrown
        """
        while not self.exception_thrown():
            try:
                return self._task_queues[worker_id].get(block=True, timeout=0.01)
            except queue.Empty:
                pass
        return None

    def task_done(self, worker_id: int) -> None:
        """
        Signal that we've completed a task

        :param worker_id: Worker ID
        """
        self._task_queues[worker_id].task_done()

    def add_results(self, worker_id: int, results: Any) -> None:
        """
        :param worker_id: Worker ID
        :param results: Results from the main function
        """
        self._results_queue.put((worker_id, results))

    def get_results(self, block: bool = True, timeout: Optional[float] = None, in_thread: bool = False) -> Any:
        """
        Obtain the next result from the results queue. We catch the queue.Empty and raise it again after the
        DelayedKeyboardInterrupt, to clean up the exception traceback whenever a keyboard interrupt is issued. If we
        wouldn't have done this, there would be an additional queue.Empty error traceback, which we don't want.

        :param block: Whether to block (wait for results)
        :param timeout: How long to wait for results in case ``block==True``
        :param in_thread: Whether this function is called from within a thread
        :return: The next result from the queue, which is the result of calling the function
        """
        queue_empty_error = None
        with DelayedKeyboardInterrupt(in_thread=in_thread):
            try:
                self._last_completed_task_worker_id, results = self._results_queue.get(block=block, timeout=timeout)
                self._results_queue.task_done()
            except queue.Empty as e:
                queue_empty_error = e
        if queue_empty_error is not None:
            raise queue_empty_error
        return results

    def add_exit_results(self, worker_id: int, results: Any) -> None:
        """
        :param worker_id: Worker ID
        :param results: Results from the exit function
        """
        self._exit_results_queues[worker_id].put(results)

    def get_exit_results(self, worker_id: int, block: bool = True, in_thread: bool = False) -> Any:
        """
        Obtain exit results from a specific worker. When block=False and the queue is empty we catch the queue.Empty and
        raise it again after the DelayedKeyboardInterrupt, to clean up the exception traceback whenever a keyboard
        interrupt is issued. If we wouldn't have done this, there would be an additional queue.Empty error traceback,
        which we don't want.

        :param worker_id: Worker ID
        :param block: Whether to block (wait for results)
        :param in_thread: Whether this function is called from within a thread
        :return: Exit results
        """
        # We always check the queue when block=False. This is needed when keep_alive=True and the workers have not been
        # explicitly stopped and joined. In that case terminate() is called from WorkerPool.__exit__ and we need to
        # drain the queue. When terminate is called the exception_thrown is set
        while not self.exception_thrown() or not block:
            queue_empty_error = None
            with DelayedKeyboardInterrupt(in_thread=in_thread):
                try:
                    results = self._exit_results_queues[worker_id].get(block=block, timeout=0.01)
                    self._exit_results_queues[worker_id].task_done()
                    return results
                except queue.Empty as e:
                    if not block:
                        queue_empty_error = e
            if queue_empty_error is not None:
                raise queue_empty_error

    def get_exit_results_all_workers(self) -> List[Any]:
        """
        Obtain results from the exit results queues (should be done before joining the workers). When an error occurred
        inside the exit function we return immediately. There's always exactly one entry in a queue.

        :return: Exit results
        """
        exit_results = []
        for worker_id in range(self.n_jobs):
            results = self.get_exit_results(worker_id)
            if self.exception_thrown():
                return exit_results
            exit_results.append(results)
        return exit_results

    def signal_all_exit_results_obtained(self) -> None:
        """
        Signal that all exit results have been obtained
        """
        self._all_exit_results_obtained.set()

    def wait_until_all_exit_results_obtained(self) -> None:
        """
        Waits until all exit results have been obtained
        """
        while not self.exception_thrown():
            if self._all_exit_results_obtained.wait(timeout=0.01):
                return

    def add_new_map_params(self, map_params: WorkerMapParams) -> None:
        """
        Submits new map params for each worker

        :param map_params: New map params
        """
        for worker_id in range(self.n_jobs):
            self.add_task(NEW_MAP_PARAMS_PILL, worker_id)
            self.add_task(map_params, worker_id)

    ################
    # Exceptions
    ################

    def add_exception(self, err_type: type, traceback_str: str) -> None:
        """
        Add exception details to the queue and set the exception event

        :param err_type: Type of the exception
        :param traceback_str: Traceback string
        """
        self._exception_queue.put((err_type, traceback_str))

    def add_exception_poison_pill(self) -> None:
        """
        Signals the exception handling thread to shut down
        """
        with DelayedKeyboardInterrupt():
            self._exception_queue.put((POISON_PILL, POISON_PILL))

    def get_exception(self, in_thread: bool = False) -> Tuple[type, str]:
        """
        :return: Tuple containing the type of the exception and the traceback string
        """
        with DelayedKeyboardInterrupt(in_thread):
            return self._exception_queue.get(block=True)

    def task_done_exception(self) -> None:
        """
        Signal that we've completed a task for the exception queue
        """
        self._exception_queue.task_done()

    def signal_exception_thrown(self) -> None:
        """
        Set the exception event
        """
        self._exception_thrown.set()

    def exception_thrown(self) -> bool:
        """
        :return: Whether an exception was thrown by one of the workers
        """
        return self._exception_thrown.is_set()

    def wait_for_exception_thrown(self, timeout: Optional[float]) -> bool:
        """
        Waits until the exception thrown event is set

        :param timeout: How long to wait before giving up
        :return: True when exception was thrown, False if timeout was reached
        """
        return self._exception_thrown.wait(timeout=timeout)

    def signal_kill_signal_received(self) -> None:
        """
        Set the kill signal received event
        """
        self._kill_signal_received.set()

    def kill_signal_received(self) -> bool:
        """
        :return: Whether a kill signal was received in one of the workers
        """
        return self._kill_signal_received.is_set()

    ################
    # Terminating & restarting
    ################

    def insert_poison_pill(self) -> None:
        """
        'Tell' the workers their job is done.
        """
        for worker_id in range(self.n_jobs):
            self.add_task(POISON_PILL, worker_id)

    def insert_non_lethal_poison_pill(self) -> None:
        """
        When ``keep_alive=True``, the workers should stay alive, but they need to wrap up their work (like sending the
        latest progress bar update)
        """
        for worker_id in range(self.n_jobs):
            self.add_task(NON_LETHAL_POISON_PILL, worker_id)

    def signal_worker_restart(self, worker_id: int) -> None:
        """
        Signal to the main process that this worker needs to be restarted

        :param worker_id: Worker ID
        """
        self._worker_done_array[worker_id] = True

    def get_worker_restarts(self) -> Generator[int, None, None]:
        """
        Obtain the worker IDs that need to be restarted.

        :return: Generator of worker IDs
        """
        return (worker_id for worker_id, restart_worker in enumerate(self._worker_done_array) if restart_worker)

    def reset_worker_restart(self, worker_id) -> None:
        """
        Worker has been restarted, reset signal.

        :param worker_id: Worker ID
        """
        self._worker_done_array[worker_id] = False

    def get_worker_dead_lock(self, worker_id: int) -> mp.Lock:
        """
        Returns the worker dead lock for a specific worker

        :param worker_id: Worker ID
        :return: Lock object
        """
        return self._workers_dead_locks[worker_id]

    def signal_worker_alive(self, worker_id: int) -> None:
        """
        Indicate that a worker is alive

        :param worker_id: Worker ID
        """
        self._workers_dead[worker_id].clear()

    def signal_worker_dead(self, worker_id: int) -> None:
        """
`       Indicate that a worker is dead

        :param worker_id: Worker ID
        """
        self._workers_dead[worker_id].set()

    def is_worker_alive(self, worker_id: int) -> bool:
        """
        Check whether the worker is alive

        :param worker_id: Worker ID
        :return: Whether the worker is alive
        """
        return not self._workers_dead[worker_id].is_set()

    def wait_for_dead_worker(self, worker_id: int, timeout=None) -> None:
        """
        Wait until a worker is dead

        :param worker_id: Worker ID
        :param timeout: How long to wait for a worker to be dead
        """
        self._workers_dead[worker_id].wait(timeout=timeout)

    def join_results_queues(self, keep_alive: bool = False) -> None:
        """
        Join results and exit results queues

        :param keep_alive: Whether to keep the queues alive
        """
        self._results_queue.join()
        [q.join() for q in self._exit_results_queues]
        if not keep_alive:
            self._results_queue.close()
            self._results_queue.join_thread()
            [q.close() for q in self._exit_results_queues]
            [q.join_thread() for q in self._exit_results_queues]

    def join_task_queues(self, keep_alive: bool = False) -> None:
        """
        Join task queues

        :param keep_alive: Whether to keep the queues alive
        """
        [q.join() for q in self._task_queues]
        if not keep_alive:
            [q.close() for q in self._task_queues]
            [q.join_thread() for q in self._task_queues]

    def join_exception_queue(self, keep_alive: bool = False) -> None:
        """
        Join exception queue

        :param keep_alive: Whether to keep the queues alive
        """
        self._exception_queue.join()
        if not keep_alive:
            self._exception_queue.close()
            self._exception_queue.join_thread()

    def drain_queues_terminate_worker(self, worker_id: int, dont_wait_event: threading.Event) -> None:
        """
        Drain the results queues without blocking. This is done when terminating workers, while they could still be busy
        putting something in the queues. This function will always be called from within a thread.

        :param worker_id: Worker ID
        :param dont_wait_event: Event object to indicate whether other termination threads should continue. I.e., when
            we set it to False, threads should wait.
        """
        # Get results from the results queue. If we got any, keep going and inform the other termination threads to wait
        # until this one's finished
        got_results = False
        try:
            while True:
                self.get_results(block=False, in_thread=True)
                dont_wait_event.clear()
                got_results = True
        except (queue.Empty, OSError):
            if got_results:
                dont_wait_event.set()

        # Get results from the exit results queue. If we got any, keep going and inform the other termination threads to
        # wait until this one's finished
        if self._exit_results_queues:
            got_results = False
            try:
                self.get_exit_results(worker_id, block=False, in_thread=True)
                dont_wait_event.clear()
                got_results = True
            except (queue.Empty, OSError):
                if got_results:
                    dont_wait_event.set()

    def drain_queues(self) -> None:
        """
        Drain tasks, results, and exit results queues. Note that the exception queue doesn't need to be drained.
        This one is properly cleaned up in the exception handling class.
        """
        [self.drain_and_join_queue(q) for q in self._task_queues]
        self.drain_and_join_queue(self._results_queue)
        [self.drain_and_join_queue(q) for q in self._exit_results_queues]

    def drain_and_join_queue(self, q: mp.JoinableQueue, join: bool = True) -> None:
        """
        Drains a queue completely, such that it is joinable. If a timeout is reached, we give up and terminate. So far,
        I've only seen it happen when an exception is thrown when using spawn as start method, and even then it only
        happens once every 1000 runs or so.

        :param q: Queue to join
        :param join: Whether to join the queue or not
        """
        try:
            process = mp.Process(target=self._drain_and_join_queue, args=(q, join))
            process.start()
            process.join(timeout=5)
            if process.is_alive():
                logger.debug("Draining queue failed, skipping")
                process.terminate()
                process.join()
        except (OSError, RuntimeError):
            # For Windows compatibility
            pass

    @staticmethod
    def _drain_and_join_queue(q: mp.JoinableQueue, join: bool = True) -> None:
        """
        Drains a queue completely, such that it is joinable

        :param q: Queue to join
        :param join: Whether to join the queue or not
        """
        # Do nothing when it's not set
        if q is None:
            return

        # Call task done up to the point where we get a ValueError. We need to do this when child processes already
        # started processing on some tasks and got terminated half-way.
        n = 0
        try:
            while True:
                q.task_done()
                n += 1
        except (OSError, ValueError):
            pass

        try:
            while not q.empty() or n != 0:
                q.get(block=True, timeout=1.0)
                n -= 1
        except (OSError, EOFError, queue.Empty):
            pass

        # Join
        if join:
            try:
                q.join()
                q.close()
                q.join_thread()
            except OSError:
                pass



import argparse
import signal
from typing import Sequence

from mpire.dashboard import start_dashboard


def get_port_range() -> Sequence:
    """
    :return: port range
    """
    def _port_range(range_str) -> Sequence:
        n1, n2 = map(int, range_str.split('-'))
        if len(range(n1, n2)) < 2:
            raise ValueError
        return range(n1, n2)

    parser = argparse.ArgumentParser(description='MPIRE Dashboard')
    parser.add_argument('--port-range', dest='port_range', required=False, default=range(8080, 8100), type=_port_range,
                        help='Port range for starting a dashboard. The range should accommodate at least two ports: '
                             'one for the webserver and one for the Python Manager server. Example: 6060-6080 will be '
                             'converted to `range(6060, 6080)`. Default: `range(8080, 8100)`.')
    return parser.parse_args().port_range


if __name__ == '__main__':
    # Obtain port range
    port_range = get_port_range()

    # Start a dashboard
    print("Starting MPIRE dashboard...")
    dashboard_details = start_dashboard(port_range)

    # Print some details on how to connect
    print()
    print("MPIRE dashboard started on http://localhost:{}".format(dashboard_details['dashboard_port_nr']))
    print("Server is listening on {}:{}".format(dashboard_details['manager_host'],
                                                dashboard_details['manager_port_nr']))
    print("-" * 50)
    signal.pause()



import logging
import sys
from datetime import datetime, timedelta
from threading import Event, Thread
from typing import Any, Dict, Optional, Type

from tqdm.auto import tqdm

from mpire.comms import WorkerComms, POISON_PILL
from mpire.dashboard.connection_utils import (DashboardConnectionDetails, get_dashboard_connection_details,
                                              set_dashboard_connection)
from mpire.insights import WorkerInsights
from mpire.params import WorkerMapParams, WorkerPoolParams
from mpire.signal import DisableKeyboardInterruptSignal
from mpire.tqdm_utils import TqdmConnectionDetails, TqdmManager
from mpire.utils import format_seconds

# If a user has not installed the dashboard dependencies than the imports below will fail
try:
    from mpire.dashboard.dashboard import DASHBOARD_STARTED_EVENT
    from mpire.dashboard.utils import get_function_details
    from mpire.dashboard.manager import get_manager_client_dicts
except ImportError:
    DASHBOARD_STARTED_EVENT = None

    def get_function_details(_):
        pass

    def get_manager_client_dicts():
        raise NotImplementedError

logger = logging.getLogger(__name__)

DATETIME_FORMAT = "%Y-%m-%d, %H:%M:%S"


class ProgressBarHandler:

    def __init__(self, pool_params: WorkerPoolParams, map_params: WorkerMapParams, show_progress_bar: bool,
                 progress_bar_total: int, progress_bar_position: int, worker_comms: WorkerComms,
                 worker_insights: WorkerInsights) -> None:
        """
        :param pool_params: WorkerPool parameters
        :param map_params: Map parameters
        :param show_progress_bar: When ``True`` will display a progress bar
        :param progress_bar_total: Total number of tasks that will be processed
        :param progress_bar_position: Denotes the position (line nr) of the progress bar. This is useful wel using
            multiple progress bars at the same time
        :param worker_comms: Worker communication objects (queues, locks, events, ...)
        :param worker_insights: WorkerInsights object which stores the worker insights
        """
        self.show_progress_bar = show_progress_bar
        self.progress_bar_total = progress_bar_total
        self.progress_bar_position = progress_bar_position
        self.worker_comms = worker_comms
        self.worker_insights = worker_insights
        if show_progress_bar and DASHBOARD_STARTED_EVENT is not None:
            self.function_details = get_function_details(map_params.func)
            self.function_details['n_jobs'] = pool_params.n_jobs
        else:
            self.function_details = None

        self.thread = None
        self.thread_started = Event()
        self.progress_bar_id = None
        self.dashboard_dict = None
        self.dashboard_details_dict = None
        self.start_t = None

    def __enter__(self) -> 'ProgressBarHandler':
        """
        Enables the use of the ``with`` statement. Starts a new progress handler thread if a progress bar should be
        shown

        :return: self
        """
        if self.show_progress_bar:

            # Disable the interrupt signal. We let the thread die gracefully
            with DisableKeyboardInterruptSignal():
                logger.debug("Starting progress bar handler")
                self.thread = Thread(target=self._progress_bar_handler, args=(TqdmManager.get_connection_details(),
                                                                              get_dashboard_connection_details()))
                self.thread.start()
                self.thread_started.wait()

        return self

    def __exit__(self, exc_type: Type, *_) -> None:
        """
        Enables the use of the ``with`` statement. Terminates the progress handler thread if there is one
        """
        if self.show_progress_bar and self.thread.is_alive():

            # If this exit is called with an exception, then we assume an external kill signal was received (this is,
            # for example, necessary in nested pools when an error occurs)
            if exc_type is not None:
                self.worker_comms.signal_kill_signal_received()

            # Signal shutdown and close the handling thread
            if not self.worker_comms.exception_thrown():
                logger.debug("Signalling progress bar to shut down")
                self.worker_comms.signal_progress_bar_shutdown()
            logger.debug("Joining progress bar handler")
            self.thread.join()
            logger.debug("Progress bar handler joined")

    def _progress_bar_handler(self, tqdm_connection_details: TqdmConnectionDetails,
                              dashboard_connection_details: DashboardConnectionDetails) -> None:
        """
        Keeps track of the progress made by the workers and updates the progress bar accordingly

        :param tqdm_connection_details: Tqdm manager host, and whether the manager is started/connected
        :param dashboard_connection_details: Dashboard manager host, port_nr and whether a dashboard is
            started/connected
        """
        logger.debug("Progress bar handler started")

        # Set tqdm and dashboard connection details. This is needed for nested pools and in the case forkserver or
        # spawn is used as start method
        TqdmManager.set_connection_details(tqdm_connection_details)
        set_dashboard_connection(dashboard_connection_details)

        # Connect to the tqdm manager
        tqdm_manager = TqdmManager()
        tqdm_lock, tqdm_position_register = tqdm_manager.get_lock_and_position_register()
        tqdm.set_lock(tqdm_lock)
        main_progress_bar = tqdm_position_register.register_progress_bar_position(self.progress_bar_position)

        # In case we're running tqdm in a notebook we need to apply a dirty hack to get progress bars working.
        # Solution adapted from https://github.com/tqdm/tqdm/issues/485#issuecomment-473338308
        try:
            in_notebook = 'IPKernelApp' in sys.modules['IPython'].get_ipython().config
        except (AttributeError, KeyError):
            in_notebook = False
        if in_notebook and not main_progress_bar:
            print(' ', end='', flush=True)

        # Create progress bar and register the start time
        tqdm.monitor_interval = False
        progress_bar = tqdm(total=self.progress_bar_total, position=self.progress_bar_position, dynamic_ncols=True,
                            leave=True, mininterval=0.1, maxinterval=0.5)
        self.start_t = datetime.fromtimestamp(progress_bar.start_t)

        # Notify that the main process can continue working. We set it after the progress bar has been created, instead
        # of right after this thread has started, for a better user experience
        self.thread_started.set()

        # Register progress bar to dashboard in case a dashboard is started
        self._register_progress_bar(progress_bar)

        while True:
            # Wait for a job to finish
            tasks_completed = self.worker_comms.get_tasks_completed_progress_bar()

            # If we received a poison pill, we should quit right away. We do force a final refresh of the progress bar
            # to show the latest status
            if tasks_completed is POISON_PILL:
                logger.debug("Terminating progress bar handler")

                # Check if we got a poison pill because there was an error. If so, we obtain the exception information
                # and send it to the dashboard, if available.)
                if self.worker_comms.exception_thrown() or self.worker_comms.kill_signal_received():
                    progress_bar.set_description('Exception occurred, terminating ... ')
                    if self.worker_comms.exception_thrown():
                        _, traceback_str = self.worker_comms.get_exception(in_thread=True)
                        self._send_dashboard_update(progress_bar, failed=True, traceback_str=traceback_str)
                        self.worker_comms.task_done_exception()
                    elif self.worker_comms.kill_signal_received():
                        self._send_dashboard_update(progress_bar, failed=True, traceback_str='Kill signal received')

                # Final update of the progress bar. When we're not in a notebook and this is the main progress bar, we
                # add as many newlines as the highest progress bar position, such that new output is added after the
                # progress bars.
                progress_bar.refresh()
                if in_notebook:
                    progress_bar.close()
                else:
                    progress_bar.disable = True
                    if main_progress_bar:
                        progress_bar.fp.write('\n' * (tqdm_position_register.get_highest_progress_bar_position() + 1))
                break

            # Check if there's an actual update
            elif tasks_completed > 0 and tasks_completed == progress_bar.n:
                continue

            # Update progress bar
            progress_bar.update(tasks_completed - progress_bar.n)

            # Force a refresh when we're at 100%
            if progress_bar.n == progress_bar.total:
                if in_notebook:
                    progress_bar.close()
                progress_bar.refresh()
                self.worker_comms.signal_progress_bar_complete()
                self.worker_comms.wait_until_progress_bar_is_complete()
                self._send_dashboard_update(progress_bar)

            # Send update to dashboard in case a dashboard is started, but only when tqdm updated its view as well. This
            # will make the dashboard a lot more responsive
            if progress_bar.n == progress_bar.last_print_n:
                self._send_dashboard_update(progress_bar)

        logger.debug("Progress bar handler done")

    def _register_progress_bar(self, progress_bar: tqdm) -> None:
        """
        Register this progress bar to the dashboard

        :param progress_bar: tqdm progress bar instance
        """
        if self.progress_bar_id is None and DASHBOARD_STARTED_EVENT is not None and DASHBOARD_STARTED_EVENT.is_set():

            # Connect to manager server
            self.dashboard_dict, self.dashboard_details_dict, dashboard_tqdm_lock = get_manager_client_dicts()

            # Register new progress bar
            logger.debug("Registering new progress bar to the dashboard server")
            dashboard_tqdm_lock.acquire()
            self.progress_bar_id = len(self.dashboard_dict.keys()) + 1
            self.dashboard_details_dict.update([(self.progress_bar_id, self.function_details)])
            self._send_dashboard_update(progress_bar)
            dashboard_tqdm_lock.release()

    def _send_dashboard_update(self, progress_bar: tqdm, failed: bool = False,
                               traceback_str: Optional[str] = None) -> None:
        """
        Adds a progress bar update to the shared dict so the dashboard process can use it, only when a dashboard has
        started

        :param progress_bar: tqdm progress bar instance
        :param failed: Whether or not the operation failed or not
        :param traceback_str: Traceback string, if an exception was raised
        """
        if self.progress_bar_id is not None:
            self.dashboard_dict.update([(self.progress_bar_id,
                                         self._get_progress_bar_update_dict(progress_bar, failed, traceback_str))])

    def _get_progress_bar_update_dict(self, progress_bar: tqdm, failed: bool,
                                      traceback_str: Optional[str] = None) -> Dict[str, Any]:
        """
        Obtain update dictionary with all the information needed for displaying on the dashboard

        :param progress_bar: tqdm progress bar instance
        :param failed: Whether or not the operation failed or not
        :param traceback_str: Traceback string, if an exception was raised
        :return: Update dictionary
        """
        # Save some variables first so we can use them consistently with the same value
        details = progress_bar.format_dict
        n = details["n"]
        total = details["total"]
        now = datetime.now()
        remaining_time = (total - n) / details["rate"] if details["rate"] else None

        return {"id": self.progress_bar_id,
                "success": not failed,
                "n": n,
                "total": total,
                "percentage": n / total,
                "duration": str(now - self.start_t).rsplit('.', 1)[0],
                "remaining": format_seconds(remaining_time, False),
                "started_raw": self.start_t,
                "started": self.start_t.strftime(DATETIME_FORMAT),
                "finished_raw": now + timedelta(seconds=remaining_time) if remaining_time is not None else None,
                "finished": ((now + timedelta(seconds=remaining_time)).strftime(DATETIME_FORMAT)
                             if remaining_time is not None else ''),
                "traceback": traceback_str.strip() if traceback_str is not None else None,
                "insights": self.worker_insights.get_insights()}



import ctypes
import multiprocessing as mp
import unittest
from datetime import datetime
from multiprocessing import managers
from time import sleep
from unittest.mock import patch

from mpire import WorkerPool
from mpire.context import DEFAULT_START_METHOD
from mpire.insights import RUNNING_WINDOWS, WorkerInsights
from tests.utils import MockDatetimeNow


def square(x):
    # time.sleep is added for Windows compatibility, otherwise it says 0.0 time has passed
    sleep(0.001)
    return x * x


class WorkerInsightsTest(unittest.TestCase):

    def test_reset_insights(self):
        """
        Test if resetting the insights is done properly
        """
        for n_jobs in [1, 2, 4]:
            insights = WorkerInsights(mp.get_context(DEFAULT_START_METHOD), n_jobs)
            self.assertEqual(insights.ctx, mp.get_context(DEFAULT_START_METHOD))
            self.assertEqual(insights.n_jobs, n_jobs)

            with self.subTest('initialized', n_jobs=n_jobs):
                self.assertFalse(insights.insights_enabled)
                self.assertIsNone(insights.insights_manager)
                self.assertIsNone(insights.insights_manager_lock)
                self.assertIsNone(insights.worker_start_up_time)
                self.assertIsNone(insights.worker_init_time)
                self.assertIsNone(insights.worker_n_completed_tasks)
                self.assertIsNone(insights.worker_waiting_time)
                self.assertIsNone(insights.worker_working_time)
                self.assertIsNone(insights.worker_exit_time)
                self.assertIsNone(insights.max_task_duration)
                self.assertIsNone(insights.max_task_args)

            # Containers should be properly initialized
            MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 1, 2, 3, 4)]
            MockDatetimeNow.CURRENT_IDX = 0
            with self.subTest('without initial values', n_jobs=n_jobs, enable_insights=True), \
                    patch('mpire.insights.datetime', new=MockDatetimeNow):
                insights.reset_insights(enable_insights=True)
                self.assertTrue(insights.insights_enabled)
                self.assertIsInstance(insights.insights_manager_lock, mp.synchronize.Lock)
                if RUNNING_WINDOWS:
                    self.assertIsNone(insights.insights_manager)
                else:
                    self.assertIsInstance(insights.insights_manager, managers.SyncManager)
                self.assertIsInstance(insights.worker_start_up_time, ctypes.Array)
                self.assertIsInstance(insights.worker_init_time, ctypes.Array)
                self.assertIsInstance(insights.worker_n_completed_tasks, ctypes.Array)
                self.assertIsInstance(insights.worker_waiting_time, ctypes.Array)
                self.assertIsInstance(insights.worker_working_time, ctypes.Array)
                self.assertIsInstance(insights.worker_exit_time, ctypes.Array)
                self.assertIsInstance(insights.max_task_duration, ctypes.Array)
                self.assertIsInstance(insights.max_task_args, list if RUNNING_WINDOWS else managers.ListProxy)

                # Basic sanity checks for the values
                self.assertEqual(sum(insights.worker_start_up_time), 0)
                self.assertEqual(sum(insights.worker_init_time), 0)
                self.assertEqual(sum(insights.worker_n_completed_tasks), 0)
                self.assertEqual(sum(insights.worker_waiting_time), 0)
                self.assertEqual(sum(insights.worker_working_time), 0)
                self.assertEqual(sum(insights.worker_exit_time), 0)
                self.assertEqual(sum(insights.max_task_duration), 0)
                if not RUNNING_WINDOWS:
                    self.assertListEqual(list(insights.max_task_args), [''] * n_jobs * 5)

            # Set some values so we can test if the containers will be properly resetted
            insights.worker_start_up_time[0] = 1
            insights.worker_init_time[0] = 2
            insights.worker_n_completed_tasks[0] = 3
            insights.worker_waiting_time[0] = 4
            insights.worker_working_time[0] = 5
            insights.worker_exit_time[0] = 6
            insights.max_task_duration[0] = 7
            insights.max_task_args[0] = '8'

            # Containers should be properly initialized
            MockDatetimeNow.CURRENT_IDX = 0
            with self.subTest('with initial values', n_jobs=n_jobs, enable_insights=True), \
                    patch('mpire.insights.datetime', new=MockDatetimeNow):
                insights.reset_insights(enable_insights=True)
                # Basic sanity checks for the values
                self.assertEqual(sum(insights.worker_start_up_time), 0)
                self.assertEqual(sum(insights.worker_init_time), 0)
                self.assertEqual(sum(insights.worker_n_completed_tasks), 0)
                self.assertEqual(sum(insights.worker_waiting_time), 0)
                self.assertEqual(sum(insights.worker_working_time), 0)
                self.assertEqual(sum(insights.worker_exit_time), 0)
                self.assertEqual(sum(insights.max_task_duration), 0)
                self.assertListEqual(list(insights.max_task_args), [''] * n_jobs * 5)

            # Disabling should set things to None again
            with self.subTest(n_jobs=n_jobs, enable_insights=False):
                insights.reset_insights(enable_insights=False)
                self.assertFalse(insights.insights_enabled)
                self.assertIsNone(insights.insights_manager)
                self.assertIsNone(insights.insights_manager_lock)
                self.assertIsNone(insights.worker_start_up_time)
                self.assertIsNone(insights.worker_init_time)
                self.assertIsNone(insights.worker_n_completed_tasks)
                self.assertIsNone(insights.worker_waiting_time)
                self.assertIsNone(insights.worker_working_time)
                self.assertIsNone(insights.worker_exit_time)
                self.assertIsNone(insights.max_task_duration)
                self.assertIsNone(insights.max_task_args)

    def test_enable_insights(self):
        """
        Insight containers are initially set to None values. When enabled they should be changed to appropriate
        containers. When a second task is started it should reset them. If disabled, they should remain None
        """
        with WorkerPool(n_jobs=2, enable_insights=True) as pool:

            # We run this a few times to see if it resets properly. We only verify this by checking the
            # n_completed_tasks
            for idx in range(3):
                with self.subTest('enabled', idx=idx):

                    pool.map(square, self._get_tasks(10), worker_init=self._init, worker_exit=self._exit)

                    # Basic sanity checks for the values. Some max task args can be empty, in that case the duration
                    # should be 0 (= no data)
                    self.assertGreater(sum(pool._worker_insights.worker_start_up_time), 0)
                    self.assertGreater(sum(pool._worker_insights.worker_init_time), 0)
                    self.assertEqual(sum(pool._worker_insights.worker_n_completed_tasks), 10)
                    self.assertGreater(sum(pool._worker_insights.worker_waiting_time), 0)
                    self.assertGreater(sum(pool._worker_insights.worker_working_time), 0)
                    self.assertGreater(sum(pool._worker_insights.worker_exit_time), 0)
                    self.assertGreater(max(pool._worker_insights.max_task_duration), 0)
                    for duration, args in zip(pool._worker_insights.max_task_duration,
                                              pool._worker_insights.max_task_args):
                        if duration == 0:
                            self.assertEqual(args, '')
                        elif not RUNNING_WINDOWS:
                            self.assertIn(args, {'Arg 0: 0', 'Arg 0: 1', 'Arg 0: 2', 'Arg 0: 3', 'Arg 0: 4',
                                                 'Arg 0: 5', 'Arg 0: 6', 'Arg 0: 7', 'Arg 0: 8', 'Arg 0: 9'})

        with WorkerPool(n_jobs=2, enable_insights=False) as pool:

            # Disabling should set things to None again
            with self.subTest('disable'):
                pool.map(square, range(10))
                self.assertIsNone(pool._worker_insights.insights_manager)
                self.assertIsNone(pool._worker_insights.insights_manager_lock)
                self.assertIsNone(pool._worker_insights.worker_start_up_time)
                self.assertIsNone(pool._worker_insights.worker_init_time)
                self.assertIsNone(pool._worker_insights.worker_n_completed_tasks)
                self.assertIsNone(pool._worker_insights.worker_waiting_time)
                self.assertIsNone(pool._worker_insights.worker_working_time)
                self.assertIsNone(pool._worker_insights.worker_exit_time)
                self.assertIsNone(pool._worker_insights.max_task_duration)
                self.assertIsNone(pool._worker_insights.max_task_args)

    def test_get_max_task_duration_list(self):
        """
        Test that the right containers are selected given a worker ID
        """
        insights = WorkerInsights(mp.get_context(DEFAULT_START_METHOD), n_jobs=5)

        with self.subTest(insights_enabled=False):
            for worker_id in range(5):
                self.assertIsNone(insights.get_max_task_duration_list(worker_id))

        with self.subTest(insights_enabled=True):
            insights.reset_insights(enable_insights=True)
            insights.max_task_duration[:] = range(25)
            insights.max_task_args[:] = map(str, range(25))

            for worker_id, expected_task_duration_list in [
                (0, [(0.0, '0'), (1.0, '1'), (2.0, '2'), (3.0, '3'), (4.0, '4')]),
                (1, [(5.0, '5'), (6.0, '6'), (7.0, '7'), (8.0, '8'), (9.0, '9')]),
                (4, [(20.0, '20'), (21.0, '21'), (22.0, '22'), (23.0, '23'), (24.0, '24')])
            ]:
                with self.subTest(worker_id=worker_id):
                    self.assertListEqual(insights.get_max_task_duration_list(worker_id), expected_task_duration_list)

    def test_update_start_up_time(self):
        """
        Test that the start up time is correctly added to worker_start_up_time for the right index
        """
        MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 2, 0),
                                         datetime(1970, 1, 1, 0, 0, 3, 0),
                                         datetime(1970, 1, 1, 0, 0, 7, 0),
                                         datetime(1970, 1, 1, 0, 0, 8, 0)]
        MockDatetimeNow.CURRENT_IDX = 0

        insights = WorkerInsights(mp.get_context(DEFAULT_START_METHOD), n_jobs=5)

        # Shouldn't do anything when insights haven't been enabled
        with self.subTest(insights_enabled=False), patch('mpire.insights.datetime', new=MockDatetimeNow):
            for worker_id in range(5):
                insights.update_start_up_time(worker_id, datetime(1970, 1, 1, 0, 0, 0, 0))
            self.assertIsNone(insights.worker_start_up_time)

        insights.reset_insights(enable_insights=True)
        MockDatetimeNow.CURRENT_IDX = 0

        with self.subTest(insights_enabled=True), patch('mpire.insights.datetime', new=MockDatetimeNow):
            for worker_id in range(5):
                insights.update_start_up_time(worker_id, datetime(1970, 1, 1, 0, 0, 0, 0))
            self.assertListEqual(list(insights.worker_start_up_time), [0, 2, 3, 7, 8])

    def test_update_n_completed_tasks(self):
        """
        Test that the number of completed tasks is correctly added to worker_n_completed_tasks for the right index
        """
        insights = WorkerInsights(mp.get_context(DEFAULT_START_METHOD), n_jobs=5)

        # Shouldn't do anything when insights haven't been enabled
        with self.subTest(insights_enabled=False):
            for worker_id, call_n_times in [(0, 1), (1, 0), (2, 5), (3, 8), (4, 11)]:
                for _ in range(call_n_times):
                    insights.update_n_completed_tasks(worker_id)
            self.assertIsNone(insights.worker_n_completed_tasks)

        with self.subTest(insights_enabled=True):
            insights.reset_insights(enable_insights=True)
            for worker_id, call_n_times in [(0, 1), (1, 0), (2, 5), (3, 8), (4, 11)]:
                for _ in range(call_n_times):
                    insights.update_n_completed_tasks(worker_id)
            self.assertListEqual(list(insights.worker_n_completed_tasks), [1, 0, 5, 8, 11])

    def test_update_task_insights_not_forced(self):
        """
        Test whether the update_task_insights is triggered correctly when ``force_update=False``.
        """
        MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 0, 0, 0),
                                         datetime(1970, 1, 1, 0, 0, 2, 0),
                                         datetime(1970, 1, 1, 0, 0, 3, 0),
                                         datetime(1970, 1, 1, 0, 0, 7, 0),
                                         datetime(1970, 1, 1, 0, 0, 8, 0)]
        MockDatetimeNow.CURRENT_IDX = 0

        insights = WorkerInsights(mp.get_context(DEFAULT_START_METHOD), n_jobs=5)
        max_task_duration_last_updated = datetime(1970, 1, 1, 0, 0, 1, 0)

        # Shouldn't do anything when insights haven't been enabled
        with self.subTest(insights_enabled=False), patch('mpire.insights.datetime', new=MockDatetimeNow):
            for worker_id in range(5):
                max_task_duration_list = insights.get_max_task_duration_list(worker_id)
                insights.update_task_insights(worker_id, max_task_duration_last_updated, max_task_duration_list,
                                              force_update=False)
            self.assertIsNone(insights.max_task_duration)
            self.assertIsNone(insights.max_task_args)

        insights.reset_insights(enable_insights=True)
        max_task_duration_last_updated = datetime(1970, 1, 1, 0, 0, 1, 0)
        MockDatetimeNow.CURRENT_IDX = 0

        # The first three worker IDs won't send an update because the two seconds hasn't passed yet.
        with self.subTest(insights_enabled=True), patch('mpire.insights.datetime', new=MockDatetimeNow):
            last_updated_times = []
            for worker_id, max_task_duration_list in [
                (0, [(0.1, '0'), (0.2, '1'), (0.3, '2'), (0.4, '3'), (0.5, '4')]),
                (1, [(1.1, '0'), (1.2, '1'), (1.3, '2'), (1.4, '3'), (1.5, '4')]),
                (2, [(2.1, '0'), (2.2, '1'), (2.3, '2'), (2.4, '3'), (2.5, '4')]),
                (3, [(3.1, '0'), (3.2, '1'), (3.3, '2'), (3.4, '3'), (3.5, '4')]),
                (4, [(4.1, '0'), (4.2, '1'), (4.3, '2'), (4.4, '3'), (4.5, '4')])
            ]:
                last_updated_times.append(insights.update_task_insights(
                    worker_id, max_task_duration_last_updated, max_task_duration_list, force_update=False
                ))
            self.assertListEqual(last_updated_times, [datetime(1970, 1, 1, 0, 0, 1), datetime(1970, 1, 1, 0, 0, 1),
                                                      datetime(1970, 1, 1, 0, 0, 1), datetime(1970, 1, 1, 0, 0, 7),
                                                      datetime(1970, 1, 1, 0, 0, 8)])
            self.assertListEqual(list(insights.max_task_duration), [0.0, 0.0, 0.0, 0.0, 0.0,
                                                                    0.0, 0.0, 0.0, 0.0, 0.0,
                                                                    0.0, 0.0, 0.0, 0.0, 0.0,
                                                                    3.1, 3.2, 3.3, 3.4, 3.5,
                                                                    4.1, 4.2, 4.3, 4.4, 4.5])
            self.assertListEqual(list(insights.max_task_args), ['', '', '', '', '',
                                                                '', '', '', '', '',
                                                                '', '', '', '', '',
                                                                '0', '1', '2', '3', '4',
                                                                '0', '1', '2', '3', '4'])

    def test_update_task_insights_forced(self):
        """
        Test whether task insights are being updated correctly
        """
        MockDatetimeNow.RETURN_VALUES = [datetime(1970, 1, 1, 0, 0, 1, 0),
                                         datetime(1970, 1, 1, 0, 0, 2, 0)]
        MockDatetimeNow.CURRENT_IDX = 0

        insights = WorkerInsights(mp.get_context(DEFAULT_START_METHOD), n_jobs=2)
        max_task_duration_last_updated = datetime(1970, 1, 1, 0, 0, 0, 0)

        # Shouldn't do anything when insights haven't been enabled
        with self.subTest(insights_enabled=False), patch('mpire.insights.datetime', new=MockDatetimeNow):
            for worker_id in range(2):
                max_task_duration_last_updated = insights.update_task_insights(
                    worker_id, max_task_duration_last_updated, [(0.1, '1'), (0.2, '2')], force_update=True
                )
            self.assertIsNone(insights.max_task_duration)
            self.assertIsNone(insights.max_task_args)
            self.assertEqual(max_task_duration_last_updated, datetime(1970, 1, 1, 0, 0, 0, 0))

        insights.reset_insights(enable_insights=True)
        max_task_duration_last_updated = datetime(1970, 1, 1, 0, 0, 0, 0)
        MockDatetimeNow.CURRENT_IDX = 0

        with self.subTest(insights_enabled=True), patch('mpire.insights.datetime', new=MockDatetimeNow):
            for worker_id, max_task_duration_list in [(0, [(5, '5'), (6, '6'), (7, '7'), (8, '8'), (9, '9')]),
                                                      (1, [(0, '0'), (1, '1'), (2, '2'), (3, '3'), (4, '4')])]:
                self.assertEqual(insights.update_task_insights(
                    worker_id, max_task_duration_last_updated, max_task_duration_list, force_update=True
                ), datetime(1970, 1, 1, 0, 0, worker_id + 1, 0))
            self.assertListEqual(list(insights.max_task_duration), [5, 6, 7, 8, 9, 0, 1, 2, 3, 4])
            self.assertListEqual(list(insights.max_task_args), ['5', '6', '7', '8', '9', '0', '1', '2', '3', '4'])

    def test_get_insights(self):
        """
        Test if the insights are properly processed
        """
        insights = WorkerInsights(mp.get_context(DEFAULT_START_METHOD), n_jobs=2)

        with self.subTest(enable_insights=False):
            insights.reset_insights(enable_insights=False)
            self.assertDictEqual(insights.get_insights(), {})

        with self.subTest(enable_insights=True):
            insights.reset_insights(enable_insights=True)
            insights.worker_start_up_time[:] = [0.1, 0.2]
            insights.worker_init_time[:] = [0.11, 0.22]
            insights.worker_n_completed_tasks[:] = [2, 3]
            insights.worker_waiting_time[:] = [0.4, 0.3]
            insights.worker_working_time[:] = [42.0, 37.0]
            insights.worker_exit_time[:] = [0.33, 0.44]

            # Durations that are zero or args that are empty are skipped
            insights.max_task_duration[:] = [0.0, 0.0, 1.0, 2.0, 0.0, 6.0, 0.8, 0.0, 0.1, 0.0]
            insights.max_task_args[:] = ['', '', '1', '2', '', '3', '4', '', '5', '']
            insights_dict = insights.get_insights()

            # Test ratios separately because of rounding errors
            total_time = 0.3 + 0.33 + 0.7 + 79.0 + 0.77
            self.assertAlmostEqual(insights_dict['start_up_ratio'], 0.3 / total_time)
            self.assertAlmostEqual(insights_dict['init_ratio'], 0.33 / total_time)
            self.assertAlmostEqual(insights_dict['waiting_ratio'], 0.7 / total_time)
            self.assertAlmostEqual(insights_dict['working_ratio'], 79.0 / total_time)
            self.assertAlmostEqual(insights_dict['exit_ratio'], 0.77 / total_time)
            del (insights_dict['start_up_ratio'], insights_dict['init_ratio'], insights_dict['waiting_ratio'],
                 insights_dict['working_ratio'], insights_dict['exit_ratio'])

            self.assertDictEqual(insights_dict, {
                'n_completed_tasks': [2, 3],
                'start_up_time': ['0:00:00.100', '0:00:00.200'],
                'init_time': ['0:00:00.110', '0:00:00.220'],
                'waiting_time': ['0:00:00.400', '0:00:00.300'],
                'working_time': ['0:00:42', '0:00:37'],
                'exit_time': ['0:00:00.330', '0:00:00.440'],
                'total_start_up_time': '0:00:00.300',
                'total_init_time': '0:00:00.330',
                'total_waiting_time': '0:00:00.700',
                'total_working_time': '0:01:19',
                'total_exit_time': '0:00:00.770',
                'top_5_max_task_durations': ['0:00:06', '0:00:02', '0:00:01', '0:00:00.800', '0:00:00.100'],
                'top_5_max_task_args': ['', '', '', '', ''] if RUNNING_WINDOWS else ['3', '2', '1', '4', '5'],
                'total_time': '0:01:21.100',
                'start_up_time_mean': '0:00:00.150', 'start_up_time_std': '0:00:00.050',
                'init_time_mean': '0:00:00.165', 'init_time_std': '0:00:00.055',
                'waiting_time_mean': '0:00:00.350', 'waiting_time_std': '0:00:00.050',
                'working_time_mean': '0:00:39.500', 'working_time_std': '0:00:02.500',
                'exit_time_mean': '0:00:00.385', 'exit_time_std': '0:00:00.055'
            })

    @staticmethod
    def _get_tasks(n):
        """ Simulate that getting tasks takes some time """
        for i in range(n):
            # sleep is added for Windows compatibility, otherwise it says 0.0 time has passed
            sleep(0.001)
            yield i

    @staticmethod
    def _init():
        # sleep is added for Windows compatibility, otherwise it says 0.0 time has passed
        sleep(0.001)

    @staticmethod
    def _exit():
        # sleep is added for Windows compatibility, otherwise it says 0.0 time has passed
        sleep(0.001)



from datetime import datetime
from typing import Callable


class MockDatetimeNow(datetime):
    """ Mocked datetime class which returns the predefined return values in order when .now() is called """
    RETURN_VALUES = []
    CURRENT_IDX = 0

    @classmethod
    def now(cls, *_):
        cls.CURRENT_IDX += 1
        return cls.RETURN_VALUES[cls.CURRENT_IDX - 1]


class ConditionalDecorator:

    def __init__(self, decorator: Callable, condition: bool) -> None:
        """
        Decorator which takes a decorator and a condition as input. Only when the condition is met the decorator is used

        :param decorator: Decorator
        :param condition: Condition (boolean)
        """
        self.decorator = decorator
        self.condition = condition

    def __call__(self, func) -> Callable:
        """
        Enables the conditional decorator

        :param func: Function to decorated
        :return: Decorated function if condition is met, otherwise just the function
        """
        if self.condition:
            return self.decorator(func)
        else:
            return func




import atexit
import errno
import getpass
import logging
import socket
from datetime import datetime
from multiprocessing import Event, Process, Value
from pkg_resources import resource_string
from typing import Dict, Optional, Sequence, Union

from flask import escape, Flask, jsonify, render_template, request
from werkzeug.serving import make_server

from mpire.signal import DisableKeyboardInterruptSignal, ignore_keyboard_interrupt
from mpire.dashboard.manager import (DASHBOARD_MANAGER_HOST, DASHBOARD_MANAGER_PORT,
                                     get_manager_client_dicts, start_manager_server)

logger = logging.getLogger(__name__)
logger_werkzeug = logging.getLogger('werkzeug')
logger_werkzeug.setLevel(logging.ERROR)
app = Flask(__name__)
_server = None
_progress_bar_html = resource_string(__name__, 'templates/progress_bar.html').decode('utf-8')

_DASHBOARD_MANAGER = None
_DASHBOARD_TQDM_DICT = None
_DASHBOARD_TQDM_DETAILS_DICT = None
DASHBOARD_STARTED_EVENT = Event()


@app.route('/')
def index() -> str:
    """
    Obtain the index HTML

    :return: HTML
    """
    return render_template('index.html', username=getpass.getuser(), hostname=socket.gethostname(),
                           manager_host=DASHBOARD_MANAGER_HOST.value.decode() or 'localhost',
                           manager_port_nr=DASHBOARD_MANAGER_PORT.value)


@app.route('/_progress_bar_update')
def progress_bar_update() -> str:
    """
    Obtain progress bar updates (should be called through AJAX)

    :return: JSON string containing progress bar updates
    """
    # As we get updates only when the progress bar is updated we need to fix the 'duration' and 'time remaining' parts
    # (time never stops)
    now = datetime.now()
    result = []
    for pb_id in sorted(_DASHBOARD_TQDM_DICT.keys()):
        progress = _DASHBOARD_TQDM_DICT.get(pb_id)
        if progress['success'] and progress['n'] != progress['total']:
            progress['duration'] = str(now - progress['started_raw']).rsplit('.', 1)[0]
            progress['remaining'] = (str(progress['finished_raw'] - now).rsplit('.', 1)[0]
                                     if progress['finished_raw'] is not None and progress['finished_raw'] > now
                                     else '-')
        result.append(progress)

    return jsonify(result=result)


@app.route('/_progress_bar_new')
def progress_bar_new() -> str:
    """
    Obtain a piece of HTML for a new progress bar (should be called through AJAX)

    :return: JSON string containing new progress bar HTML
    """
    pb_id = int(request.args['pb_id'])
    has_insights = request.args['has_insights'] == 'true'

    # Obtain progress bar details. Only show the user@host part if it doesn't equal the user@host of this process
    # (in case someone connected to this dashboard from another machine or user)
    progress_bar_details = _DASHBOARD_TQDM_DETAILS_DICT.get(pb_id)
    if progress_bar_details['user'] == '{}@{}'.format(getpass.getuser(), socket.gethostname()):
        progress_bar_details['user'] = ''
    else:
        progress_bar_details['user'] = '{}:'.format(progress_bar_details['user'])

    # Create table for worker insights
    insights_workers = []
    if has_insights:
        for worker_id in range(progress_bar_details['n_jobs']):
            insights_workers.append(f"<tr><td>{worker_id}</td>"
                                    f"<td id='pb_{pb_id}_insights_worker_{worker_id}_tasks_completed'></td>"
                                    f"<td id='pb_{pb_id}_insights_worker_{worker_id}_start_up_time'></td>"
                                    f"<td id='pb_{pb_id}_insights_worker_{worker_id}_init_time'></td>"
                                    f"<td id='pb_{pb_id}_insights_worker_{worker_id}_waiting_time'></td>"
                                    f"<td id='pb_{pb_id}_insights_worker_{worker_id}_working_time'></td>"
                                    f"<td id='pb_{pb_id}_insights_worker_{worker_id}_exit_time'></td>"
                                    f"</tr>")
    insights_workers = "\n".join(insights_workers)

    return jsonify(result=_progress_bar_html.format(id=pb_id, insights_workers=insights_workers,
                                                    has_insights='block' if has_insights else 'none',
                                                    **{k: escape(v) for k, v in progress_bar_details.items()}))


def start_dashboard(port_range: Sequence = range(8080, 8100)) -> Dict[str, Union[int, str]]:
    """
    Starts a new MPIRE dashboard

    :param port_range: Port range to try.
    :return: A dictionary containing the dashboard port number and manager host and port_nr being used
    """
    global _DASHBOARD_MANAGER, _DASHBOARD_TQDM_DICT, _DASHBOARD_TQDM_DETAILS_DICT

    if not DASHBOARD_STARTED_EVENT.is_set():

        # Prevent signal from propagating to child process
        with DisableKeyboardInterruptSignal():

            # Set up manager server
            _DASHBOARD_MANAGER = start_manager_server(port_range)

            # Start flask server
            logging.getLogger('werkzeug').setLevel(logging.WARN)
            dashboard_port_nr = Value('i', 0, lock=False)
            Process(target=_run, args=(DASHBOARD_STARTED_EVENT, DASHBOARD_MANAGER_HOST.value,
                                       DASHBOARD_MANAGER_PORT.value, dashboard_port_nr, port_range),
                    daemon=True, name='dashboard-process').start()
            DASHBOARD_STARTED_EVENT.wait()

            # Return connect information
            return {'dashboard_port_nr': dashboard_port_nr.value,
                    'manager_host': DASHBOARD_MANAGER_HOST.value.decode() or socket.gethostname(),
                    'manager_port_nr': DASHBOARD_MANAGER_PORT.value}

    else:
        raise RuntimeError("You already have a running dashboard")


def connect_to_dashboard(manager_port_nr: int, manager_host: Optional[Union[bytes, str]] = None) -> None:
    """
    Connects to an existing MPIRE dashboard

    :param manager_port_nr: Port to use when connecting to a manager
    :param manager_host: Host to use when connecting to a manager. If ``None`` it will use localhost
    """
    global _DASHBOARD_MANAGER, _DASHBOARD_TQDM_DICT, _DASHBOARD_TQDM_DETAILS_DICT

    if not DASHBOARD_STARTED_EVENT.is_set():
        # Set connection variables so we can connect to the right manager
        manager_host = manager_host or "127.0.0.1"
        if isinstance(manager_host, str):
            manager_host = manager_host.encode()
        DASHBOARD_MANAGER_HOST.value = manager_host
        DASHBOARD_MANAGER_PORT.value = manager_port_nr
        DASHBOARD_STARTED_EVENT.set()
    else:
        raise RuntimeError("You're already connected to a running dashboard")


def _run(started: Event, manager_host: str, manager_port_nr: int, dashboard_port_nr: Value,
         port_range: Sequence) -> None:
    """
    Starts a dashboard server

    :param started: Event that signals the dashboard server has started
    :param manager_host: Dashboard manager host
    :param manager_port_nr: Dashboard manager port number
    :param dashboard_port_nr: Value object for storing the dashboad port number that is used
    :param port_range: Port range to try.
    """
    ignore_keyboard_interrupt()  # For Windows compatibility

    # Set dashboard connection details. This is needed when spawn is the default start method
    DASHBOARD_MANAGER_HOST.value = manager_host
    DASHBOARD_MANAGER_PORT.value = manager_port_nr

    # Connect to manager from this process
    global _DASHBOARD_TQDM_DICT, _DASHBOARD_TQDM_DETAILS_DICT, _server
    _DASHBOARD_TQDM_DICT, _DASHBOARD_TQDM_DETAILS_DICT, _ = get_manager_client_dicts()

    # Try different ports, until a free one is found
    for port in port_range:
        try:
            _server = make_server('0.0.0.0', port, app)
            dashboard_port_nr.value = port
            started.set()
            logger.info("Server started on 0.0.0.0:%d", port)
            _server.serve_forever()
            break
        except OSError as exc:
            if exc.errno != errno.EADDRINUSE:
                raise exc

    if not _server:
        raise OSError(f"Dashboard server: All ports are in use: {port_range}")


@atexit.register
def stop() -> None:
    """
    Called when the program exits, will shutdown the server
    """
    global _server
    if _server:
        try:
            _server.shutdown()
        except Exception:
            pass



from typing import Optional, Tuple

# If a user has not installed the dashboard dependencies than the imports below will fail
try:
    from mpire.dashboard import connect_to_dashboard
    from mpire.dashboard.dashboard import DASHBOARD_STARTED_EVENT
    from mpire.dashboard.manager import DASHBOARD_MANAGER_HOST, DASHBOARD_MANAGER_PORT
except (ImportError, ModuleNotFoundError):
    DASHBOARD_MANAGER_HOST = None
    DASHBOARD_MANAGER_PORT = None
    DASHBOARD_STARTED_EVENT = None

    def connect_to_dashboard(*_):
        pass

DashboardConnectionDetails = Tuple[Optional[bytes], Optional[int], bool]


def get_dashboard_connection_details() -> DashboardConnectionDetails:
    """
    Obtains the connection details of a dasbhoard. These details are needed to be passed on to child process when the
    start method is either forkserver or spawn.

    :return: Dashboard manager host, port_nr and whether a dashboard is started/connected
    """
    return (DASHBOARD_MANAGER_HOST.value if DASHBOARD_MANAGER_HOST is not None else None,
            DASHBOARD_MANAGER_PORT.value if DASHBOARD_MANAGER_PORT is not None else None,
            DASHBOARD_STARTED_EVENT.is_set() if DASHBOARD_STARTED_EVENT is not None else False)


def set_dashboard_connection(dashboard_connection_details: DashboardConnectionDetails,
                             auto_connect: bool = True) -> None:
    """
    Sets the dashboard connection details and connects to an existing dashboard if needed.

    :param dashboard_connection_details:  Dashboard manager host, port_nr and whether a dashboard is started/connected
    :param auto_connect: Whether to automatically connect to a sever when the dashboard_started event is set
    """
    dashboard_manager_host, dashboard_manager_port_nr, dashboard_started = dashboard_connection_details
    if (dashboard_manager_host is not None and dashboard_manager_port_nr is not None and
            not DASHBOARD_STARTED_EVENT.is_set()):
        if dashboard_started and auto_connect:
            connect_to_dashboard(dashboard_manager_port_nr, dashboard_manager_host)
        else:
            DASHBOARD_MANAGER_HOST.value = dashboard_manager_host
            DASHBOARD_MANAGER_PORT.value = dashboard_manager_port_nr
            if dashboard_started:
                DASHBOARD_STARTED_EVENT.set()
